import pytest
import torch

from curated_transformers.layers.attention import (
    _TORCH_SDP,
    AttentionLinearBiases,
    AttentionMask,
    enable_torch_sdp,
)
from curated_transformers.models.bert.encoder import BERTEncoder
from curated_transformers.models.gpt_neox.decoder import GPTNeoXDecoder

from ..compat import has_hf_transformers
from ..conftest import TORCH_DEVICES
from ..utils import torch_assertclose

N_PIECES = 1024


def test_context_manager():
    # Check default.
    assert not _TORCH_SDP.get()

    # Check context manager default.
    with enable_torch_sdp():
        assert _TORCH_SDP.get()

    # Check that the original value is restored.
    assert not _TORCH_SDP.get()

    # Check with explicit argument.
    with enable_torch_sdp(True):
        assert _TORCH_SDP.get()
    with enable_torch_sdp(False):
        assert not _TORCH_SDP.get()


@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.parametrize("torch_device", TORCH_DEVICES)
def test_torch_sdp(torch_device):
    model = BERTEncoder.from_hf_hub(
        name="explosion-testing/bert-test", device=torch_device
    )
    model.eval()

    torch.manual_seed(0)
    X = torch.randint(0, N_PIECES, (2, 10), device=torch_device)
    mask = torch.ones_like(X, dtype=torch.bool)
    with torch.no_grad():
        Y = model(X, AttentionMask(mask)).last_hidden_layer_state
        with enable_torch_sdp():
            Y_sdp = model(X, AttentionMask(mask)).last_hidden_layer_state
    torch_assertclose(Y, Y_sdp)


@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.parametrize("torch_device", TORCH_DEVICES)
def test_torch_sdp_mask(torch_device):
    model = BERTEncoder.from_hf_hub(
        name="explosion-testing/bert-test", device=torch_device
    )
    model.eval()

    torch.manual_seed(0)
    X = torch.randint(0, N_PIECES, (2, 10), device=torch_device)
    mask = torch.rand((2, 10), dtype=torch.float, device=torch_device) < 0.5
    with torch.no_grad():
        Y = model(
            X, attention_mask=AttentionMask(mask)
        ).last_hidden_layer_state * mask.unsqueeze(-1)
        with enable_torch_sdp():
            Y_sdp = model(
                X, attention_mask=AttentionMask(mask)
            ).last_hidden_layer_state * mask.unsqueeze(-1)
    torch_assertclose(Y, Y_sdp)


@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.parametrize("torch_device", TORCH_DEVICES)
def test_torch_sdp_causal(torch_device):
    model = GPTNeoXDecoder.from_hf_hub(
        name="trl-internal-testing/tiny-random-GPTNeoXForCausalLM", device=torch_device
    )
    model.eval()

    torch.manual_seed(0)
    X = torch.randint(0, N_PIECES, (2, 10), device=torch_device)
    mask = torch.ones_like(X, dtype=torch.bool)
    with torch.no_grad():
        Y = model(X, AttentionMask(mask)).last_hidden_layer_state
        with enable_torch_sdp():
            Y_sdp = model(X, AttentionMask(mask)).last_hidden_layer_state
    torch_assertclose(Y, Y_sdp)


@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
@pytest.mark.parametrize("torch_device", TORCH_DEVICES)
def test_torch_sdp_causal_with_mask(torch_device):
    model = GPTNeoXDecoder.from_hf_hub(
        name="trl-internal-testing/tiny-random-GPTNeoXForCausalLM", device=torch_device
    )
    model.eval()

    torch.manual_seed(0)
    X = torch.randint(0, N_PIECES, (2, 10), device=torch_device)
    mask = torch.rand((2, 10), dtype=torch.float, device=torch_device) < 0.5
    with torch.no_grad():
        Y = model(
            X, attention_mask=AttentionMask(mask)
        ).last_hidden_layer_state * mask.unsqueeze(-1)
        with enable_torch_sdp():
            Y_sdp = model(
                X, attention_mask=AttentionMask(mask)
            ).last_hidden_layer_state * mask.unsqueeze(-1)
    torch_assertclose(Y, Y_sdp)


@pytest.mark.parametrize("torch_device", TORCH_DEVICES)
def test_attention_linear_biases(torch_device):
    pow2_slopes = AttentionLinearBiases(
        n_attention_heads=8, is_causal=False, is_inverted=False
    ).slopes
    torch_assertclose(
        pow2_slopes.to(device=torch_device),
        torch.tensor(
            [
                [
                    [[0.5]],
                    [[0.25]],
                    [[0.125]],
                    [[0.0625]],
                    [[0.03125]],
                    [[0.015625]],
                    [[0.0078125]],
                    [[0.00390625]],
                ]
            ],
            device=torch_device,
        ),
    )
    non_pow2_slopes = AttentionLinearBiases(
        n_attention_heads=12, is_causal=False, is_inverted=False
    ).slopes
    torch_assertclose(
        non_pow2_slopes.to(device=torch_device),
        torch.tensor(
            [
                [
                    [[0.5]],
                    [[0.25]],
                    [[0.125]],
                    [[0.0625]],
                    [[0.03125]],
                    [[0.015625]],
                    [[0.0078125]],
                    [[0.00390625]],
                    [[0.7071067811865476]],
                    [[0.35355339059327384]],
                    [[0.17677669529663692]],
                    [[0.08838834764831849]],
                ]
            ],
            device=torch_device,
        ),
    )

    alibi_causal = AttentionLinearBiases(
        n_attention_heads=4, is_causal=True, is_inverted=False
    )
    torch_assertclose(
        alibi_causal(attention_scores=torch.zeros((1, 4, 1, 3), device=torch_device)),
        torch.tensor(
            [
                [
                    [[-0.5000, -0.2500, 0.0000]],
                    [[-0.1250, -0.0625, 0.0000]],
                    [[-0.03125, -0.015625, 0.0000]],
                    [[-0.0078125, -0.00390625, 0.0000]],
                ]
            ],
            device=torch_device,
        ),
    )

    alibi_non_causal = AttentionLinearBiases(
        n_attention_heads=4, is_causal=False, is_inverted=False
    )
    torch_assertclose(
        alibi_non_causal(
            attention_scores=torch.zeros((1, 4, 3, 3), device=torch_device)
        ),
        torch.tensor(
            [
                [
                    [
                        [0.0000, -0.2500, -0.5000],
                        [-0.2500, 0.0000, -0.2500],
                        [-0.5000, -0.2500, 0.0000],
                    ],
                    [
                        [0.0000, -0.0625, -0.1250],
                        [-0.0625, 0.0000, -0.0625],
                        [-0.1250, -0.0625, 0.0000],
                    ],
                    [
                        [0.0000, -0.015625, -0.03125],
                        [-0.015625, 0.0000, -0.015625],
                        [-0.03125, -0.015625, 0.0000],
                    ],
                    [
                        [0.0000, -0.00390625, -0.0078125],
                        [-0.00390625, 0.0000, -0.00390625],
                        [-0.0078125, -0.00390625, 0.0000],
                    ],
                ]
            ],
            device=torch_device,
        ),
    )

    alibi_causal_inverted = AttentionLinearBiases(
        n_attention_heads=4, is_causal=True, is_inverted=True
    )
    torch_assertclose(
        alibi_causal_inverted(
            attention_scores=torch.zeros((1, 4, 1, 3), device=torch_device)
        ),
        torch.tensor(
            [
                [
                    [[0.0000, 0.2500, 0.5000]],
                    [[0.0000, 0.0625, 0.1250]],
                    [[0.0000, 0.015625, 0.03125]],
                    [[0.0000, 0.00390625, 0.0078125]],
                ]
            ],
            device=torch_device,
        ),
    )

    alibi_non_causal_inverted = AttentionLinearBiases(
        n_attention_heads=4, is_causal=False, is_inverted=True
    )
    torch_assertclose(
        alibi_non_causal_inverted(
            attention_scores=torch.zeros((1, 4, 3, 3), device=torch_device)
        ),
        torch.tensor(
            [
                [
                    [
                        [0.5000, 0.2500, 0.0000],
                        [0.2500, 0.5000, 0.2500],
                        [0.0000, 0.2500, 0.5000],
                    ],
                    [
                        [0.1250, 0.0625, 0.0000],
                        [0.0625, 0.1250, 0.0625],
                        [0.0000, 0.0625, 0.1250],
                    ],
                    [
                        [0.03125, 0.015625, 0.0000],
                        [0.015625, 0.03125, 0.015625],
                        [0.0000, 0.015625, 0.03125],
                    ],
                    [
                        [0.0078125, 0.00390625, 0.0000],
                        [0.00390625, 0.0078125, 0.00390625],
                        [0.0000, 0.00390625, 0.0078125],
                    ],
                ]
            ],
            device=torch_device,
        ),
    )



from typing import Any, Dict, Mapping, Type, TypeVar

from torch import Tensor

from ..bert import BERTConfig as ELECTRAConfig
from ..bert import BERTEncoder
from ..hf_hub.conversion import state_dict_from_hf, state_dict_to_hf
from ._hf import HF_PARAM_KEY_TRANSFORMS, _config_from_hf, _config_to_hf

# Only provided as typing.Self in Python 3.11+.
Self = TypeVar("Self", bound="ELECTRAEncoder")


class ELECTRAEncoder(BERTEncoder):
    """
    ELECTRA (`Clark et al., 2020`_) encoder.

    .. _Clark et al., 2020 : https://arxiv.org/abs/2003.10555
    """

    @classmethod
    def is_supported(cls: Type[Self], config: Dict[str, Any]) -> bool:
        return config.get("model_type") == "electra"

    @classmethod
    def state_dict_from_hf(
        cls: Type[Self], params: Mapping[str, Tensor]
    ) -> Mapping[str, Tensor]:
        return state_dict_from_hf(params, HF_PARAM_KEY_TRANSFORMS)

    @classmethod
    def state_dict_to_hf(
        cls: Type[Self], params: Mapping[str, Tensor]
    ) -> Mapping[str, Tensor]:
        return state_dict_to_hf(params, HF_PARAM_KEY_TRANSFORMS)

    @classmethod
    def config_from_hf(cls, hf_config: Mapping[str, Any]) -> ELECTRAConfig:
        return _config_from_hf(hf_config)

    @classmethod
    def config_to_hf(cls, curated_config: ELECTRAConfig) -> Mapping[str, Any]:
        return _config_to_hf(curated_config)



from typing import Any, Generic, Iterator, List, Optional, Tuple

import torch
from torch import Tensor
from torch.distributions import Categorical

from ..layers.attention import AttentionMask
from ..models.module import CausalLMModule
from ..models.output import CacheT, CausalLMOutputWithCache
from .config import GeneratorConfig, GreedyGeneratorConfig, SampleGeneratorConfig
from .logits import LogitsTransform
from .state import GeneratorState


class Generator(Generic[CacheT]):
    """
    Generator base class for causal language models.
    """

    model: CausalLMModule[Any, CacheT]

    def __init__(self, model: CausalLMModule[Any, CacheT]):
        """
        Construct a generator.

        :param model:
            The causal language model to generate with.
        """
        self.model = model

    def __call__(
        self,
        *,
        attention_mask: AttentionMask,
        ids: Tensor,
        config: GeneratorConfig,
    ) -> Iterator[Tuple[Tensor, Tensor]]:
        """
        Alias for :meth:`.generate`.
        """
        return self.generate(
            attention_mask=attention_mask,
            ids=ids,
            config=config,
        )

    def generate(
        self,
        *,
        attention_mask: AttentionMask,
        ids: Tensor,
        config: GeneratorConfig,
    ) -> Iterator[Tuple[Tensor, Tensor]]:
        """
        Generate text, starting from the given piece identifiers.

        The generator returns an iterator over tuples. Each tuple contains:
         1. A tensor with sequence identifiers.
         2. A tensor with the next piece identifiers.

        The sequence identifiers are numbered ``0..batch`` and are
        necessary because some sequences may finish generation earliers than
        others. The sequence identifiers allow the caller to map the generated
        pieces back to the original input sequences.

        :param ids:
            Batch of piece identifiers to start generating from.

            *Shape:* ``(batch_size, seq_len)``
        :param attention_mask:
            Attention mask that masks out pieces that should not be attended to.
        :param config:
            Generator configuraton.
        :returns:
            An iterator over tuples. Each tuple contains a tensor with the
            sequence identifiers and a tensor with the next piece identier.

            *Shape:* ``(batch_unfinished,)``
        """
        self.model.eval()

        logits_transform = config.logits_transform()
        stop_condition = config.stop_condition()
        if isinstance(config, GreedyGeneratorConfig):
            generation_step = self._decode_greedy
        elif isinstance(config, SampleGeneratorConfig):
            generation_step = self._decode_sample
        else:
            raise ValueError(
                f"Unknown generator configuration: {type(config).__name__}"
            )

        cache: Optional[List[CacheT]] = None
        state = GeneratorState(
            attention_mask=attention_mask, cache=cache, prompt_ids=ids
        )

        while True:
            with torch.no_grad():
                output = self.model(
                    state.last_step_ids,
                    attention_mask=state.attention_mask,
                    cache=state.cache,
                    store_cache=True,
                    positions=state.positions,
                )

            seq_ids, last_step_ids = state.step(
                cache=output.cache,
                predicted_ids=generation_step(logits_transform, output),
                stop_condition=stop_condition,
            )

            if seq_ids.size(0) > 0:
                yield seq_ids, last_step_ids

            if state.is_finished:
                return

    def _decode_greedy(
        self,
        logits_transform: LogitsTransform,
        output: CausalLMOutputWithCache[CacheT],
    ) -> Tensor:
        logits = logits_transform(output.logits[:, -1:, :], inplace=True)
        return logits.argmax(-1)

    def _decode_sample(
        self, logits_transform: LogitsTransform, output: CausalLMOutputWithCache[CacheT]
    ) -> Tensor:
        logits = logits_transform(output.logits[:, -1:, :], inplace=True)
        distribution = Categorical(logits=logits)
        return distribution.sample()



from copy import deepcopy

import pytest
import torch

from curated_transformers._compat import has_bitsandbytes
from curated_transformers.generation.config import GreedyGeneratorConfig
from curated_transformers.generation.dolly_v2 import DollyV2Generator
from curated_transformers.quantization.bnb import BitsAndBytesConfig

from ..conftest import GPU_TESTS_ENABLED

PROMPTS_AND_KEYWORDS = [
    [
        "What is the Rust programming language?",
        "What is spaCy?",
    ],
    [
        ["Rust", "multi-paradigm", "language", "design"],
        ["SpaCy", "NLP", "library", "Python"],
    ],
]


@pytest.fixture(scope="module")
def dolly_generator_8_bit():
    return DollyV2Generator.from_hf_hub(
        name="databricks/dolly-v2-3b",
        device=torch.device("cuda", index=0),
        quantization_config=BitsAndBytesConfig.for_8bit(),
    )


@pytest.fixture(scope="module")
def dolly_generator_4_bit():
    return DollyV2Generator.from_hf_hub(
        name="databricks/dolly-v2-3b",
        device=torch.device("cuda", index=0),
        quantization_config=BitsAndBytesConfig.for_4bit(),
    )


def _check_quantized_generator_output(output, expected_keywords):
    # Due the inherent non-determinism of executing low-bit quantized modules on
    # different GPUs (which can use different kernels), we can't reliably expect the
    # output to match a string verbatim. So, we'll just look for specific, low-frequency
    # keywords as a way to detect if gibberish/irrelevant text is being generated.
    for output, keywords in zip(output, expected_keywords):
        assert all(keyword in output for keyword in keywords)


@pytest.mark.slow
@pytest.mark.skipif(not GPU_TESTS_ENABLED, reason="requires GPU")
@pytest.mark.skipif(not has_bitsandbytes, reason="requires bitsandbytes")
def test_quantization_8_bit(dolly_generator_8_bit):
    prompts = deepcopy(PROMPTS_AND_KEYWORDS[0])
    expected = deepcopy(PROMPTS_AND_KEYWORDS[1])
    generated = dolly_generator_8_bit(
        prompts, config=GreedyGeneratorConfig(max_generated_pieces=50)
    )
    _check_quantized_generator_output(generated, expected)

    prompts = prompts[::-1]
    expected = expected[::-1]
    generated = dolly_generator_8_bit(
        prompts, config=GreedyGeneratorConfig(max_generated_pieces=50)
    )
    _check_quantized_generator_output(generated, expected)


@pytest.mark.slow
@pytest.mark.skipif(not GPU_TESTS_ENABLED, reason="requires GPU")
@pytest.mark.skipif(not has_bitsandbytes, reason="requires bitsandbytes")
def test_quantization_4_bit(dolly_generator_4_bit):
    prompts = deepcopy(PROMPTS_AND_KEYWORDS[0])
    expected = deepcopy(PROMPTS_AND_KEYWORDS[1])
    generated = dolly_generator_4_bit(
        prompts, config=GreedyGeneratorConfig(max_generated_pieces=50)
    )
    _check_quantized_generator_output(generated, expected)

    prompts = prompts[::-1]
    expected = expected[::-1]
    generated = dolly_generator_4_bit(
        prompts, config=GreedyGeneratorConfig(max_generated_pieces=50)
    )
    _check_quantized_generator_output(generated, expected)


@pytest.mark.slow
@pytest.mark.skipif(not has_bitsandbytes, reason="requires bitsandbytes")
def test_quantization_on_non_gpu():
    with pytest.raises(ValueError, match="only be performed on CUDA"):
        model = DollyV2Generator.from_hf_hub(
            name="databricks/dolly-v2-3b",
            device=None,
            quantization_config=BitsAndBytesConfig.for_8bit(),
        )

    with pytest.raises(ValueError, match="only be performed on CUDA"):
        model = DollyV2Generator.from_hf_hub(
            name="databricks/dolly-v2-3b",
            device=torch.device("cpu"),
            quantization_config=BitsAndBytesConfig.for_4bit(),
        )



from typing import Dict, Iterable, List, Optional

from curated_tokenizers import WordPieceProcessor

from ..chunks import MergedInputChunks, MergedSpecialPieceChunk
from ..tokenizer import PiecesWithIds
from .legacy_tokenizer import LegacyTokenizer


class WordPieceTokenizer(LegacyTokenizer):
    """
    Piece tokenizer using WordPiece tokenization (`Devlin et al., 2018`_).

    .. _Devlin et al., 2018 : https://arxiv.org/abs/1810.04805
    """

    def __init__(
        self,
        *,
        vocab: Dict[str, int],
        special_pieces: Optional[Dict[str, int]],
    ):
        """
        Construct a tokenizer from a ``curated-tokenizers`` WordPiece processor.

        :param vocab:
            The word piece vocabulary.
        :param special_pieces:
            Special pieces.
        """
        # Special tokens are usually already in the wordpiece vocabs, but lets
        # just add them to be sure.
        self.special_piece_to_id = {} if special_pieces is None else special_pieces
        self.id_to_special_piece = {v: k for k, v in self.special_piece_to_id.items()}
        vocab.update(self.special_piece_to_id)

        # We'll build up the vocab, verifying that the user provided ids for
        # all tokens as a sanity check.
        n_pieces = max(vocab.values()) + 1
        pieces: List[Optional[str]] = [None] * n_pieces
        for piece, idx in vocab.items():
            pieces[idx] = piece

        unused_indices = [str(id) for id, piece in enumerate(pieces) if piece is None]
        if unused_indices:
            raise ValueError(
                f"WordPiece vocabulary contains unused indices: {', '.join(unused_indices)}"
            )

        self.processor = WordPieceProcessor(pieces)

    def piece_to_id(self, piece: str) -> Optional[int]:
        return self.processor.piece_to_id(piece)

    def _decode(
        self, input: Iterable[Iterable[int]], skip_special_pieces: bool
    ) -> List[str]:
        decoded = []
        for piece_ids in input:
            tokens = []
            for piece_id in piece_ids:
                if skip_special_pieces and piece_id in self.id_to_special_piece:
                    continue

                token, is_initial = self.processor.id_to_piece(piece_id)
                if is_initial:
                    token = " " + token
                tokens.append(token)
            decoded.append("".join(tokens))
        return decoded

    def _encode(self, input: Iterable[MergedInputChunks]) -> PiecesWithIds:
        ids = []
        pieces = []

        for seq in input:
            seq_ids = []
            seq_pieces = []

            for chunk in seq:
                if isinstance(chunk, MergedSpecialPieceChunk):
                    seq_ids.append(self.processor.get_initial(chunk.piece))
                    seq_pieces.append(chunk.piece)
                else:
                    for token in chunk.text.split(" "):
                        token_ids, token_pieces = self.processor.encode(token)
                        seq_ids.extend(token_ids)
                        seq_pieces.extend(token_pieces)

            ids.append(seq_ids)
            pieces.append(seq_pieces)

        return PiecesWithIds(ids=ids, pieces=pieces)



from typing import TYPE_CHECKING, Callable, Optional, Set, Union

import torch
from torch import Tensor
from torch.nn import Module, Parameter

from ..._compat import has_bitsandbytes
from ...util.pytorch import ModuleIterator, apply_to_module
from ...util.serde.load import TensorToParameterConverterT
from .config import BitsAndBytesConfig, _4BitConfig, _8BitConfig

if TYPE_CHECKING:
    import bitsandbytes as bnb
else:
    bnb = None


def prepare_for_quantization(
    module: Module,
    config: BitsAndBytesConfig,
    non_quantizable_module_prefixes: Set[str],
) -> Optional[TensorToParameterConverterT]:
    """
    Prepares a PyTorch module for quantization using the ``bitsandbytes``
    library.

    :param module:
        Module to prepare for quantization.
    :param config:
        ``bitsandbytes`` quantization configuration.
    :param non_quantizable_module_prefixes:
        Set of module prefixes that should not be quantized.
    :returns:
        An optional callback for converting non-quantized tensors
        to parameters that are compatible with the ``bitsandbytes``
        quantization backend.
    """
    _assert_bitsandbytes_installed()

    # All inputs to quantized layers need to be of type `float16`.
    # TODO How do we exclude non-quantizable modules from this?
    module.to(torch.float16)

    if isinstance(config.inner, _8BitConfig):
        _replace_quantizable_modules(
            module, config.inner, non_quantizable_module_prefixes, _init_8bit_linear
        )
    elif isinstance(config.inner, _4BitConfig):
        _replace_quantizable_modules(
            module, config.inner, non_quantizable_module_prefixes, _init_4bit_linear
        )
    else:
        raise ValueError(f"Unknown bitsandbytes config `{type(config)}`")

    return _convert_tensor_to_quantized_parameter


def _replace_quantizable_modules(
    module: Module,
    config: Union[_8BitConfig, _4BitConfig],
    non_quantizable_module_prefixes: Set[str],
    init_quantized_module: Callable[
        [Module, Union[_8BitConfig, _4BitConfig], torch.device], Module
    ],
):
    def apply(itr: ModuleIterator):
        quantize_everything = len(non_quantizable_module_prefixes) == 0
        if not quantize_everything and itr.prefix in non_quantizable_module_prefixes:
            # Not a quantizable module.
            return
        elif itr.parent is None:
            # Root module
            return
        elif not isinstance(itr.module, torch.nn.Linear):
            # Only Linear modules can be quantized.
            return

        # Initialize on the original device, which should be the
        # `meta` device at this stage.
        device = next(itr.module.parameters()).device
        quantized_module = init_quantized_module(itr.module, config, device)
        itr.parent._modules[itr.name] = quantized_module

    apply_to_module(module, apply)


def _convert_tensor_to_quantized_parameter(
    module: Module,
    module_prefix: str,
    parameter_name: str,
    tensor: Tensor,
    device: Optional[torch.device] = None,
) -> Parameter:
    if device is None or "cuda" not in device.type:
        raise ValueError(
            f"bitsandbytes quantization can only be performed on CUDA GPU devices"
        )

    old_param = module._parameters[parameter_name]
    assert old_param is not None
    if old_param.shape != tensor.shape:
        raise ValueError(
            f"Expected size of replacement for `{module_prefix}` to be {old_param.shape}, but got {tensor.shape}"
        )

    import bitsandbytes as bnb

    # Bias is stored as a regular, non-quantized parameter.
    is_bias_param = parameter_name == "bias"
    # All parameters need to be of dtype `float16` for quantization.
    tensor = tensor.to(torch.float16)

    is_8bit = isinstance(module, bnb.nn.Linear8bitLt)
    is_4bit = isinstance(module, bnb.nn.Linear4bit)
    is_non_quantized = not is_8bit and not is_4bit

    if is_bias_param or is_non_quantized:
        new_param = Parameter(tensor, requires_grad=old_param.requires_grad)
    elif is_8bit:
        assert isinstance(old_param, bnb.nn.Int8Params)
        new_param = bnb.nn.Int8Params(
            tensor, requires_grad=False, has_fp16_weights=old_param.has_fp16_weights
        )
    elif is_4bit:
        assert isinstance(old_param, bnb.nn.Params4bit)
        new_param = bnb.nn.Params4bit(
            tensor,
            requires_grad=False,
            blocksize=old_param.blocksize,
            compress_statistics=old_param.compress_statistics,
            quant_type=old_param.quant_type,
            quant_state=old_param.quant_state,
        )

    return new_param.to(device=device)  # type:ignore


def _init_8bit_linear(
    source: Module, config: Union[_8BitConfig, _4BitConfig], device: torch.device
) -> "bnb.nn.Linear8bitLt":
    assert isinstance(config, _8BitConfig)
    import bitsandbytes as bnb

    quantized_module = bnb.nn.Linear8bitLt(
        input_features=source.in_features,
        output_features=source.out_features,
        bias=source.bias is not None,
        has_fp16_weights=config.finetunable,
        threshold=config.outlier_threshold,
        device=device,
    )
    return quantized_module


def _init_4bit_linear(
    source: Module, config: Union[_8BitConfig, _4BitConfig], device: torch.device
) -> "bnb.nn.Linear4bit":
    assert isinstance(config, _4BitConfig)
    import bitsandbytes as bnb

    quantized_module = bnb.nn.Linear4bit(
        input_features=source.in_features,
        output_features=source.out_features,
        bias=source.bias is not None,
        compute_dtype=config.compute_dtype,
        compress_statistics=config.double_quantization,
        quant_type=config.quantization_dtype.value,
        device=device,
    )
    return quantized_module


def _assert_bitsandbytes_installed():
    if not has_bitsandbytes:
        raise ValueError(
            "The `bitsandbytes` Python library is required for quantization support"
        )



import unicodedata
from typing import Any, Dict, Iterable, List, Mapping, Optional, Type, TypeVar

from curated_tokenizers import WordPieceProcessor

from ...repository.file import RepositoryFile
from .._hf_compat import clean_up_decoded_string_like_hf, tokenize_chinese_chars_bert
from ..chunks import (
    InputChunks,
    MergedInputChunks,
    MergedSpecialPieceChunk,
    SpecialPieceChunk,
    TextChunk,
)
from ..hf_hub import LegacyFromHF
from ..tokenizer import PiecesWithIds
from ..util import remove_pieces_from_sequence
from .legacy_tokenizer import (
    DefaultNormalizer,
    Normalizer,
    PostDecoder,
    PreDecoder,
    PreEncoder,
)
from .wordpiece_tokenizer import WordPieceTokenizer

# Only provided as typing.Self in Python 3.11+.
Self = TypeVar("Self", bound="BERTTokenizer")

TOKENIZER_CONFIG_MAPPING: Dict[str, str] = {
    "do_lower_case": "lowercase",
    "strip_accents": "strip_accents",
    "cls_token": "bos_piece",
    "sep_token": "eos_piece",
    "unk_token": "unk_piece",
}


class BERTPreEncoder(PreEncoder):
    def __init__(
        self,
        *,
        bos_piece: str,
        eos_piece: str,
    ):
        """
        Construct a BERT pre-encoder.

        :param bos_piece:
            The piece used to mark the beginning of a sequence.
        :param eos_piece:
            The piece used to mark the end of a sequence.
        """
        self.bos_piece = bos_piece
        self.eos_piece = eos_piece

    def split_token_on_punctuation(self, token: str) -> List[str]:
        """
        Split a token on punctuation characters. For instance,
        'AWO-Mitarbeiter' is split into ['AWO', '-', 'Mitarbeiter'].
        """
        tokens = []
        in_word = False
        while token:
            char = token[0]
            token = token[1:]
            if self.is_punctuation(char):
                tokens.append([char])
                in_word = False
            else:
                if in_word:
                    tokens[-1].append(char)
                else:
                    tokens.append([char])
                    in_word = True
        return ["".join(t) for t in tokens]

    def is_punctuation(self, char: str) -> bool:
        """
        Checks whether `char` is a punctuation character.
        """
        # ASCII punctuation from HF tranformers, since we need to split
        # in the same way.
        cp = ord(char)
        if (
            (cp >= 33 and cp <= 47)
            or (cp >= 58 and cp <= 64)
            or (cp >= 91 and cp <= 96)
            or (cp >= 123 and cp <= 126)
        ):
            return True

        return unicodedata.category(char).startswith("P")

    def __call__(self, input: Iterable[InputChunks]) -> List[InputChunks]:
        preprocessed = []

        for seq in input:
            processed_seq = InputChunks([SpecialPieceChunk(self.bos_piece)])
            for chunk in seq:
                if isinstance(chunk, TextChunk):
                    words = []
                    for word in chunk.text.split(" "):
                        word_with_punct = self.split_token_on_punctuation(word)
                        words.extend(word_with_punct)
                    processed_seq.append(TextChunk(" ".join(words)))
                else:
                    processed_seq.append(chunk)
            processed_seq.append(SpecialPieceChunk(self.eos_piece))
            preprocessed.append(processed_seq)

        return preprocessed


class BERTPreDecoder(PreDecoder):
    def __init__(
        self,
        *,
        bos_id: int,
        eos_id: int,
    ):
        """
        Construct a BERT pre-decoder.

        :param bos_id:
            The piece id used to mark the beginning of a sequence.
        :param eos_id:
            The piece id used to mark the end of a sequence.
        """
        self.bos_id = bos_id
        self.eos_id = eos_id

    def __call__(self, input: Iterable[Iterable[int]]) -> List[List[int]]:
        return [
            list(remove_pieces_from_sequence(ids, (self.bos_id, self.eos_id)))
            for ids in input
        ]


class BERTPostDecoder(PostDecoder):
    def __init__(
        self,
    ):
        """
        Construct a BERT post-decoder.
        """
        pass

    def __call__(self, output: Iterable[str]) -> List[str]:
        # The transformations done in the pre-encoding stage are lossy/non-reversible,
        # we can only do a selected number of transformations to massage the output
        # to look similar to the input text.
        return [clean_up_decoded_string_like_hf(string.strip()) for string in output]


class BERTNormalizer(Normalizer):
    """
    Performs BERT normalization operations on input chunks before encoding.
    """

    def __init__(
        self,
        *,
        lowercase: bool = False,
        strip_accents: bool = False,
        tokenize_chinese_chars: bool = True,
    ):
        """
        Construct a default normalizer.

        :param lowercase:
            Lowercase text.
        :param strip_accents:
            Remove accents from text.
        :param tokenize_chinese_chars:
            Tokenize Chinese characters.
        """

        self.default_normalizer = DefaultNormalizer(
            lowercase=lowercase, strip_accents=strip_accents
        )
        self.tokenize_chinese_chars = tokenize_chinese_chars

    def __call__(self, chunks: Iterable[InputChunks]) -> List[InputChunks]:
        chunks = self.default_normalizer(chunks)
        if self.tokenize_chinese_chars:
            for chunk in chunks:
                for piece_or_text in chunk:
                    if not isinstance(piece_or_text, TextChunk):
                        continue
                    piece_or_text.text = tokenize_chinese_chars_bert(piece_or_text.text)
        return chunks


class BERTTokenizer(WordPieceTokenizer, LegacyFromHF):
    """
    Legacy tokenizer for BERT (`Devlin et al., 2018`_) models.

    .. _Devlin et al., 2018 : https://arxiv.org/abs/1810.04805
    """

    vocab_files: Dict[str, str] = {"vocab": "vocab.txt"}

    def __init__(
        self,
        *,
        vocab: Dict[str, int],
        special_pieces: Optional[Dict[str, int]] = None,
        bos_piece: str = "[CLS]",
        eos_piece: str = "[SEP]",
        unk_piece: str = "[UNK]",
        lowercase: bool = False,
        strip_accents: bool = False,
        tokenize_chinese_chars: bool = True,
    ):
        """
        Construct a BERT tokenizer from a ``curated-tokenizers`` WordPiece processor.

        :param vocab:
            The word piece vocabulary.
        :param special_pieces:
            Special pieces.
        :param bos_piece:
            The piece used to mark the beginning of a sequence.
        :param eos_piece:
            The piece used to mark the end of a sequence.
        :param unk_piece:
            The piece used to mark unknown tokens.
        :param lowercase:
            Lowercase text.
        :param strip_accents:
            Strip accents from text.
        :param tokenize_chinese_chars:
            Tokenize Chinese characters.
        """
        super().__init__(vocab=vocab, special_pieces=special_pieces)

        self.normalizer = BERTNormalizer(
            lowercase=lowercase,
            strip_accents=strip_accents,
            tokenize_chinese_chars=tokenize_chinese_chars,
        )

        bos_id = _get_piece_id_or_fail(self.processor, bos_piece)
        eos_id = _get_piece_id_or_fail(self.processor, eos_piece)
        unk_id = _get_piece_id_or_fail(self.processor, unk_piece)

        self.unk_id = unk_id
        self.unk_piece = unk_piece
        self._eos_piece = eos_piece

        self.pre_encoder = BERTPreEncoder(
            bos_piece=bos_piece,
            eos_piece=eos_piece,
        )
        self.pre_decoder = BERTPreDecoder(bos_id=bos_id, eos_id=eos_id)
        self.post_decoder = BERTPostDecoder()

    @classmethod
    def from_files(
        cls: Type[Self],
        *,
        vocab_file: RepositoryFile,
        bos_piece: str = "[CLS]",
        eos_piece: str = "[SEP]",
        unk_piece: str = "[UNK]",
        lowercase: bool = False,
        strip_accents: bool = False,
    ) -> Self:
        """
        Construct a tokenizer from the vocabulary file.

        :param vocab_file:
            The vocabulary file.
        :param bos_piece:
            The piece to use to mark the beginning of a sequence.
        :param eos_piece:
            The piece to use to mark the end of a sequence.
        :param unk_piece:
            The piece used to mark unknown tokens.
        :param lowercase:
            Lowercase text.
        :param strip_accents:
            Strip accents from text.
        """
        vocab: Dict[str, int] = {}
        with vocab_file.open(mode="r", encoding="utf8") as f:
            for line in f:
                vocab[line.strip()] = len(vocab)

        # Add the standard special pieces to the special pieces dict.
        special_pieces = {
            piece: vocab[piece] for piece in [bos_piece, eos_piece, unk_piece]
        }

        return cls(
            vocab=vocab,
            bos_piece=bos_piece,
            eos_piece=eos_piece,
            unk_piece=unk_piece,
            lowercase=lowercase,
            strip_accents=strip_accents,
            special_pieces=special_pieces,
        )

    @property
    def eos_piece(self) -> Optional[str]:
        return self._eos_piece

    @classmethod
    def _load_from_vocab_files(
        cls: Type[Self],
        *,
        vocab_files: Mapping[str, RepositoryFile],
        tokenizer_config: Optional[Dict[str, Any]],
    ) -> Self:
        extra_kwargs = {}
        if tokenizer_config is not None:
            for hf_key, curated_key in TOKENIZER_CONFIG_MAPPING.items():
                value = tokenizer_config.get(hf_key)
                if value is not None:
                    extra_kwargs[curated_key] = value
            if "strip_accents" in extra_kwargs:
                strip_accents = extra_kwargs["strip_accents"]
                lowercase = extra_kwargs.get("lowercase", False)
                # Huggingface BERT also strips accents when lowercasing is enabled
                # and accent stripping is not defined.
                extra_kwargs["strip_accents"] = strip_accents or (
                    strip_accents is not False and lowercase
                )

        return cls.from_files(vocab_file=vocab_files["vocab"], **extra_kwargs)

    def _encode(self, input: Iterable[MergedInputChunks]) -> PiecesWithIds:
        ids = []
        pieces = []

        for seq in input:
            seq_ids = []
            seq_pieces = []

            for chunk in seq:
                if isinstance(chunk, MergedSpecialPieceChunk):
                    seq_ids.append(self.processor.get_initial(chunk.piece))
                    seq_pieces.append(chunk.piece)
                else:
                    for token in chunk.text.split(" "):
                        token_ids, token_pieces = self.processor.encode(token)
                        # Replace tokens with an unknown piece by a single
                        # unknown piece. Ideally, we'd use the pieces that
                        # we found, but this is how existing BERT models
                        # treat unknowns, so we should probably do the same.
                        if any(token_id == -1 for token_id in token_ids):
                            token_ids = [self.unk_id]
                            token_pieces = [self.unk_piece]
                        seq_ids.extend(token_ids)
                        seq_pieces.extend(token_pieces)

            ids.append(seq_ids)
            pieces.append(seq_pieces)

        return PiecesWithIds(ids=ids, pieces=pieces)


def _get_piece_id_or_fail(processor: WordPieceProcessor, piece: str):
    try:
        return processor.get_initial(piece)
    except KeyError:
        raise ValueError(
            f"BERT piece encoder vocabulary doesn't contain '{piece}' piece"
        )



import pytest
import torch

from curated_transformers.tokenizers import PiecesWithIds
from curated_transformers.tokenizers.chunks import (
    InputChunks,
    SpecialPieceChunk,
    TextChunk,
)
from curated_transformers.tokenizers.legacy.bert_tokenizer import (
    BERTPreEncoder,
    BERTTokenizer,
)
from curated_transformers.tokenizers.legacy.legacy_tokenizer import DefaultNormalizer

from ...compat import has_hf_transformers
from ...utils import torch_assertclose
from ..util import compare_tokenizer_outputs_with_hf_tokenizer


@pytest.fixture
def toy_tokenizer_from_files(test_dir):
    return BERTTokenizer.from_files(
        vocab_file=test_dir / "toy.wordpieces",
    )


@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_from_hf_hub_equals_hf_tokenizer(sample_texts):
    compare_tokenizer_outputs_with_hf_tokenizer(
        sample_texts, "bert-base-cased", BERTTokenizer
    )


@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_from_hf_hub_equals_hf_tokenizer_short(short_sample_texts):
    compare_tokenizer_outputs_with_hf_tokenizer(
        short_sample_texts, "bert-base-cased", BERTTokenizer
    )


def test_from_files(toy_tokenizer_from_files, short_sample_texts):
    encoding = toy_tokenizer_from_files(short_sample_texts)
    _check_toy_tokenizer(encoding)

    assert toy_tokenizer_from_files.decode(encoding.ids) == [
        "I saw a girl with a telescope.",
        "Today we will eat bowl, lots of it!",
        "Tokens which are unknown alphabet vocabularies.",
    ]

    assert toy_tokenizer_from_files.decode(encoding.ids, skip_special_pieces=False) == [
        "I saw a girl with a telescope.",
        "Today we will eat [UNK] bowl, lots of it!",
        "Tokens which are unknown [UNK] [UNK] [UNK] alphabet [UNK] vocabularies.",
    ]


def _check_toy_tokenizer(pieces):
    assert isinstance(pieces, PiecesWithIds)
    assert len(pieces.ids) == 3
    assert len(pieces.pieces) == 3

    assert pieces.ids == [
        [2, 41, 818, 61, 67, 193, 88, 204, 61, 251, 909, 682, 102, 95, 17, 3],
        [
            2,
            824,
            98,
            189,
            311,
            417,
            65,
            155,
            1,
            416,
            117,
            88,
            15,
            844,
            91,
            100,
            163,
            183,
            5,
            3,
        ],
        [
            2,
            576,
            159,
            100,
            365,
            319,
            356,
            99,
            93,
            281,
            1,
            1,
            1,
            340,
            102,
            103,
            608,
            184,
            1,
            809,
            90,
            608,
            328,
            162,
            742,
            17,
            3,
        ],
    ]
    assert pieces.pieces == [
        [
            "[CLS]",
            "I",
            "saw",
            "a",
            "g",
            "##ir",
            "##l",
            "with",
            "a",
            "te",
            "##les",
            "##co",
            "##p",
            "##e",
            ".",
            "[SEP]",
        ],
        [
            "[CLS]",
            "To",
            "##d",
            "##ay",
            "we",
            "will",
            "e",
            "##at",
            "[UNK]",
            "bo",
            "##w",
            "##l",
            ",",
            "lo",
            "##t",
            "##s",
            "of",
            "it",
            "!",
            "[SEP]",
        ],
        [
            "[CLS]",
            "Tok",
            "##en",
            "##s",
            "which",
            "are",
            "un",
            "##k",
            "##n",
            "##own",
            "[UNK]",
            "[UNK]",
            "[UNK]",
            "al",
            "##p",
            "##h",
            "##ab",
            "##et",
            "[UNK]",
            "vo",
            "##c",
            "##ab",
            "##ul",
            "##ar",
            "##ies",
            ".",
            "[SEP]",
        ],
    ]

    torch_assertclose(
        pieces.padded_tensor(padding_id=1),
        torch.tensor(
            [
                [
                    2,
                    41,
                    818,
                    61,
                    67,
                    193,
                    88,
                    204,
                    61,
                    251,
                    909,
                    682,
                    102,
                    95,
                    17,
                    3,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                ],
                [
                    2,
                    824,
                    98,
                    189,
                    311,
                    417,
                    65,
                    155,
                    1,
                    416,
                    117,
                    88,
                    15,
                    844,
                    91,
                    100,
                    163,
                    183,
                    5,
                    3,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                ],
                [
                    2,
                    576,
                    159,
                    100,
                    365,
                    319,
                    356,
                    99,
                    93,
                    281,
                    1,
                    1,
                    1,
                    340,
                    102,
                    103,
                    608,
                    184,
                    1,
                    809,
                    90,
                    608,
                    328,
                    162,
                    742,
                    17,
                    3,
                ],
            ],
            dtype=torch.int32,
        ),
    )
    torch_assertclose(
        pieces.attention_mask().bool_mask.squeeze(dim=(1, 2)),
        torch.tensor(
            [
                [
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                ],
                [
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                ],
                [
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                ],
            ]
        ),
    )


def test_bert_tokenizer_normalizer_preencoder():
    normalizer = DefaultNormalizer(lowercase=False, strip_accents=True)
    preencoder = BERTPreEncoder(bos_piece="[CLS]", eos_piece="[SEP]")

    def apply(input):
        return preencoder(normalizer(input))

    assert apply([InputChunks([TextChunk("AWO-Mitarbeiter")])]) == [
        InputChunks(
            [
                SpecialPieceChunk("[CLS]"),
                TextChunk("AWO - Mitarbeiter"),
                SpecialPieceChunk("[SEP]"),
            ]
        )
    ]
    assert apply([InputChunks([TextChunk("-Mitarbeiter")])]) == [
        InputChunks(
            [
                SpecialPieceChunk("[CLS]"),
                TextChunk("- Mitarbeiter"),
                SpecialPieceChunk("[SEP]"),
            ]
        )
    ]
    assert apply([InputChunks([TextChunk("AWO-")])]) == [
        InputChunks(
            [SpecialPieceChunk("[CLS]"), TextChunk("AWO -"), SpecialPieceChunk("[SEP]")]
        )
    ]
    assert apply([InputChunks([TextChunk("-")])]) == [
        InputChunks(
            [SpecialPieceChunk("[CLS]"), TextChunk("-"), SpecialPieceChunk("[SEP]")]
        )
    ]
    assert apply([InputChunks([TextChunk("")])]) == [
        InputChunks(
            [
                SpecialPieceChunk("[CLS]"),
                TextChunk(""),
                SpecialPieceChunk("[SEP]"),
            ]
        )
    ]
    assert apply([InputChunks([TextChunk("Brötchen")])]) == [
        InputChunks(
            [
                SpecialPieceChunk("[CLS]"),
                TextChunk("Brotchen"),
                SpecialPieceChunk("[SEP]"),
            ]
        )
    ]
    assert apply([InputChunks([TextChunk("Mw.-St.")])]) == [
        InputChunks(
            [
                SpecialPieceChunk("[CLS]"),
                TextChunk("Mw . - St ."),
                SpecialPieceChunk("[SEP]"),
            ]
        )
    ]



import json
from abc import ABC, abstractmethod
from json import JSONDecodeError
from typing import Any, Dict, List, Optional, Tuple

from .._compat import has_safetensors
from ..util.serde.checkpoint import ModelCheckpointType
from ._hf import (
    HF_MODEL_CONFIG,
    HF_TOKENIZER_CONFIG,
    PRIMARY_CHECKPOINT_FILENAMES,
    SHARDED_CHECKPOINT_INDEX_FILENAMES,
    SHARDED_CHECKPOINT_INDEX_WEIGHTS_KEY,
    SPECIAL_TOKENS_MAP,
    TOKENIZER_JSON,
)
from .file import RepositoryFile
from .transaction import TransactionContext


class Repository(ABC):
    """
    A repository that contains a model or tokenizer.
    """

    @abstractmethod
    def file(self, path: str) -> RepositoryFile:
        """
        Get a lazily-loaded repository file.

        :param path:
            The path of the file within the repository.
        :returns:
            The file.
        """
        ...

    def json_file(self, path: str) -> Dict[str, Any]:
        """
        Get and parse a JSON file.

        :param path:
            The path of the file within the repository.
        :returns:
            The deserialized JSON.
        :raises FileNotFoundError:
            When the file cannot be found.
        :raises OSError:
            When the file cannot be opened.
        :raises json.JSONDecodeError:
            When the JSON cannot be decoded.
        """
        file = self.file(path)
        with file.open("r", encoding="utf-8") as f:
            return json.load(f)

    @abstractmethod
    def pretty_path(self, path: Optional[str] = None) -> str:
        """
        Get a user-consumable path representation (e.g. for error messages).

        :param path:
            The path of a file within the repository. The repository path
            will be returned if ``path`` is falsy.
        :returns:
            The path representation.
        """
        ...

    @abstractmethod
    def transaction(self) -> TransactionContext:
        """
        Begins a new transaction. File operations performed on the transaction
        context will be deferred until the transaction completes successfully.

        :returns:
            The transaction context manager.
        """
        ...


class ModelRepository(Repository):
    """
    Repository wrapper that exposes some methods that are useful for working
    with repositories that contain a model.
    """

    # Cached model configuration.
    _model_config: Optional[Dict[str, Any]]

    def __init__(self, repo: Repository):
        """
        Construct a model repository wrapper.

        :param repo:
            The repository to wrap.
        """
        super().__init__()
        self.repo = repo
        self._model_config = None

    def file(self, filename: str) -> RepositoryFile:
        return self.repo.file(filename)

    def json_file(self, path: str) -> Dict[str, Any]:
        return self.repo.json_file(path)

    def model_checkpoints(self) -> Tuple[List[RepositoryFile], ModelCheckpointType]:
        """
        Retrieve the model checkpoints and checkpoint type.

        :returns:
            A tuple of the model checkpoints and the checkpoint type.
        :raises OSError:
            When the checkpoint paths files could not be retrieved.
        """

        def get_checkpoint_paths(
            checkpoint_type: ModelCheckpointType,
        ) -> List[RepositoryFile]:
            # Attempt to download a non-sharded checkpoint first.
            primary_ckpt = self.file(PRIMARY_CHECKPOINT_FILENAMES[checkpoint_type])
            if primary_ckpt.exists():
                return [primary_ckpt]

            # Try sharded checkpoint.
            index_filename = SHARDED_CHECKPOINT_INDEX_FILENAMES[checkpoint_type]

            try:
                index = self.json_file(index_filename)
            except (FileNotFoundError, JSONDecodeError, OSError) as e:
                raise OSError(
                    f"Index file for sharded checkpoint type {checkpoint_type.pretty_name} "
                    f"could not be loaded `{self.pretty_path(index_filename)}`"
                ) from e

            weight_map = index.get(SHARDED_CHECKPOINT_INDEX_WEIGHTS_KEY)
            if not isinstance(weight_map, dict):
                raise OSError(
                    f"Invalid index file in sharded {checkpoint_type.pretty_name} "
                    f"checkpoint for model at `{self.pretty_path()}`"
                )

            checkpoint_paths = []
            for filename in sorted(set(weight_map.values())):
                ckpt = self.file(filename)
                if not ckpt.exists():
                    raise OSError(
                        f"File for sharded checkpoint type {checkpoint_type.pretty_name} "
                        f"could not be found at `{self.pretty_path(index_filename)}`"
                    )
                checkpoint_paths.append(ckpt)

            return checkpoint_paths

        # Precedence: Safetensors > PyTorch
        checkpoint_type = ModelCheckpointType.SAFE_TENSORS
        checkpoint_paths: Optional[List[RepositoryFile]] = None
        if has_safetensors:
            try:
                checkpoint_paths = get_checkpoint_paths(checkpoint_type)
            except OSError:
                pass
        if checkpoint_paths is None:
            checkpoint_type = ModelCheckpointType.PYTORCH_STATE_DICT
            checkpoint_paths = get_checkpoint_paths(checkpoint_type)

        assert checkpoint_paths is not None
        assert checkpoint_type is not None
        return checkpoint_paths, checkpoint_type

    def model_config(self) -> Dict[str, Any]:
        """
        Get the model configuration. The result is cached to speed up
        subsequent lookups.

        :returns:
            The model configuration.
        :raises OSError:
            When the model config cannot be opened.
        """
        if self._model_config is None:
            try:
                self._model_config = self.json_file(
                    path=HF_MODEL_CONFIG,
                )
            except (FileNotFoundError, JSONDecodeError, OSError) as e:
                raise OSError(
                    "Cannot load config for the model at "
                    f"`{self.pretty_path(HF_MODEL_CONFIG)}`"
                ) from e

        return self._model_config

    def model_type(self) -> str:
        """
        Get the model type.

        :returns:
            The model type.
        :raises OSError:
            When the model config cannot be opened.
        """
        return self.model_config()["model_type"]

    def pretty_path(self, path: Optional[str] = None) -> str:
        return self.repo.pretty_path(path)

    def transaction(self) -> TransactionContext:
        return self.repo.transaction()


class TokenizerRepository(Repository):
    """
    Repository wrapper that exposes some methods that are useful for working
    with repositories that contain a tokenizer.
    """

    _tokenizer_config: Optional[Dict[str, Any]]

    def __init__(self, repo: Repository):
        """
        Construct a tokenizer repository wrapper.

        :param repo:
            The repository to wrap.
        """
        super().__init__()
        self.repo = repo
        self._tokenizer_config = None

    def file(self, path: str) -> RepositoryFile:
        return self.repo.file(path)

    def json_file(self, path: str) -> Dict[str, Any]:
        return self.repo.json_file(path)

    def model_type(self) -> str:
        """
        Get the model type.

        :returns:
            The model type.
        :raises OSError:
            When the model config cannot be opened.
        """
        model_config = self.json_file(HF_MODEL_CONFIG)
        return model_config["model_type"]

    def pretty_path(self, path: Optional[str] = None) -> str:
        return self.repo.pretty_path(path)

    def special_tokens_map(self) -> Dict[str, Any]:
        """
        Return the tokenizer's special tokens map.

        :returns:
            The special tokens map.
        :raises OSError:
            When the special tokens map cannot be opened.
        """
        try:
            return self.repo.json_file(
                path=SPECIAL_TOKENS_MAP,
            )
        except (FileNotFoundError, JSONDecodeError, OSError) as e:
            raise OSError(
                "Could not load special tokens map for the tokenizer at "
                f"`{self.repo.pretty_path(SPECIAL_TOKENS_MAP)}`"
            ) from e

    def tokenizer_config(self) -> Dict[str, Any]:
        """
        Return the model's tokenizer configuration. The result is cached to
        speed up subsequent lookups.

        :returns:
            Model configuration.
        :raises OSError:
            When the tokenizer config cannot be opened.
        """
        if self._tokenizer_config is None:
            try:
                self._tokenizer_config = self.repo.json_file(
                    path=HF_TOKENIZER_CONFIG,
                )
            except (FileNotFoundError, JSONDecodeError, OSError) as e:
                raise OSError(
                    "Couldn't find a valid config for the tokenizer at "
                    f"`{self.repo.pretty_path(HF_TOKENIZER_CONFIG)}`"
                ) from e

        return self._tokenizer_config

    def tokenizer_json(self) -> RepositoryFile:
        """
        Return the HF tokenizers' ``tokenizer.json``.

        :returns:
            The tokenizer file.
        """
        return self.repo.file(TOKENIZER_JSON)

    def transaction(self) -> TransactionContext:
        return self.repo.transaction()



from .albert import ALBERTConfig, ALBERTEncoder
from .auto_model import AutoCausalLM, AutoDecoder, AutoEncoder
from .bert import BERTConfig, BERTEncoder
from .camembert import CamemBERTEncoder
from .config import (
    RotaryEmbeddingConfig,
    TransformerAttentionLayerConfig,
    TransformerConfig,
    TransformerEmbeddingLayerConfig,
    TransformerFeedForwardLayerConfig,
    TransformerLayerConfig,
)
from .falcon import FalconCausalLM, FalconConfig, FalconDecoder
from .gpt_neox import GPTNeoXCausalLM, GPTNeoXConfig, GPTNeoXDecoder
from .hf_hub import FromHF
from .llama import LlamaCausalLM, LlamaConfig, LlamaDecoder
from .module import CausalLMModule, DecoderModule, EncoderModule
from .mpt import MPTCausalLM, MPTConfig, MPTDecoder
from .output import CausalLMOutputWithCache, ModelOutput, ModelOutputWithCache
from .roberta import RoBERTaConfig, RoBERTaEncoder
from .transformer import TransformerCausalLM, TransformerDecoder, TransformerEncoder
from .xlm_roberta import XLMREncoder

__all__ = [
    "ALBERTConfig",
    "ALBERTEncoder",
    "AutoCausalLM",
    "AutoDecoder",
    "AutoEncoder",
    "BERTConfig",
    "BERTEncoder",
    "CamemBERTEncoder",
    "CausalLMModule",
    "CausalLMOutputWithCache",
    "DecoderModule",
    "EncoderModule",
    "FalconCausalLM",
    "FalconConfig",
    "FalconDecoder",
    "FromHF",
    "GPTNeoXCausalLM",
    "GPTNeoXConfig",
    "GPTNeoXDecoder",
    "LlamaCausalLM",
    "LlamaConfig",
    "LlamaDecoder",
    "MPTCausalLM",
    "MPTConfig",
    "MPTDecoder",
    "ModelOutput",
    "ModelOutputWithCache",
    "RoBERTaConfig",
    "RoBERTaEncoder",
    "RotaryEmbeddingConfig",
    "TransformerAttentionLayerConfig",
    "TransformerConfig",
    "TransformerCausalLM",
    "TransformerDecoder",
    "TransformerEmbeddingLayerConfig",
    "TransformerEncoder",
    "TransformerFeedForwardLayerConfig",
    "TransformerLayerConfig",
    "XLMREncoder",
]



from functools import partial
from typing import Any, Dict, Mapping, Optional, Tuple, Type, TypeVar

import torch
from torch import Tensor
from torch.nn import Dropout, LayerNorm, ModuleList

from ...layers.attention import (
    AttentionHeads,
    QkvMode,
    QkvSplitGroupedByKVHeads,
    ScaledDotProductAttention,
    SelfAttention,
)
from ...layers.embeddings import QueryKeyRotaryEmbeddings
from ...layers.feedforward import PointwiseFeedForward
from ...layers.transformer import (
    DecoderLayer,
    EmbeddingDropouts,
    EmbeddingLayerNorms,
    TransformerDropouts,
    TransformerEmbeddings,
    TransformerLayerNorms,
)
from ..hf_hub import FromHF
from ..hf_hub.conversion import state_dict_from_hf, state_dict_to_hf
from ..transformer import TransformerDecoder
from ._hf import DECODER_HF_PARAM_KEY_TRANSFORMS, _config_from_hf, _config_to_hf
from .config import GPTNeoXConfig

# Only provided as typing.Self in Python 3.11+.
Self = TypeVar("Self", bound="GPTNeoXDecoder")


class GPTNeoXDecoder(TransformerDecoder[GPTNeoXConfig], FromHF):
    """
    GPT-NeoX (`Black et al., 2022`_) decoder.

    .. _Black et al., 2022: https://arxiv.org/abs/2204.06745
    """

    def __init__(
        self, config: GPTNeoXConfig, *, device: Optional[torch.device] = None
    ) -> None:
        """
        Construct a GPT-NeoX decoder.

        :param config:
            Decoder configuration.
        :param device:
            Device to which the module is to be moved.
        :returns:
            The decoder.
        """
        super().__init__(config)

        self.embeddings = TransformerEmbeddings(
            dropouts=EmbeddingDropouts(
                embed_output_dropout=Dropout(config.embedding.dropout_prob)
            ),
            embedding_width=config.embedding.embedding_width,
            hidden_width=config.layer.feedforward.hidden_width,
            layer_norms=EmbeddingLayerNorms(),
            n_pieces=config.embedding.n_pieces,
            n_positions=None,
            n_types=None,
            device=device,
        )

        hidden_width = config.layer.feedforward.hidden_width
        n_attention_heads = config.layer.attention.n_query_heads
        layer_norm = partial(
            LayerNorm,
            hidden_width,
            config.layer.layer_norm_eps,
            device=device,
        )
        if config.layer.attention.rotary_embeddings is None:
            raise ValueError(
                "GPT-NeoX attention config does not contain rotary embedding parameters"
            )
        self.layers = ModuleList(
            [
                DecoderLayer(
                    attention_layer=SelfAttention(
                        attention_heads=AttentionHeads.uniform(
                            n_attention_heads,
                            QkvSplitGroupedByKVHeads(),
                        ),
                        attention_scorer=ScaledDotProductAttention(
                            dropout_prob=config.layer.attention.dropout_prob,
                            linear_biases=None,
                        ),
                        hidden_width=hidden_width,
                        qkv_mode=QkvMode.MERGED_SPLIT_BEFORE,
                        rotary_embeds=QueryKeyRotaryEmbeddings(
                            fraction=config.layer.attention.rotary_embeddings.rotary_fraction,
                            base=config.layer.attention.rotary_embeddings.rotary_base,
                            head_width=hidden_width // n_attention_heads,
                        ),
                        use_bias=config.layer.attention.use_bias,
                        device=device,
                    ),
                    feed_forward_layer=PointwiseFeedForward(
                        activation=config.layer.feedforward.activation.module(),
                        hidden_width=hidden_width,
                        intermediate_width=config.layer.feedforward.intermediate_width,
                        use_bias=config.layer.feedforward.use_bias,
                        use_gate=config.layer.feedforward.use_gate,
                        device=device,
                    ),
                    dropouts=TransformerDropouts.layer_output_dropouts(
                        config.layer.dropout_prob
                    ),
                    layer_norms=TransformerLayerNorms(
                        attn_input_layer_norm=layer_norm(),
                        ffn_input_layer_norm=layer_norm(),
                    ),
                    use_parallel_attention=config.layer.attention.use_parallel_attention,
                )
                for _ in range(config.layer.n_hidden_layers)
            ]
        )

        self.output_layer_norm = LayerNorm(
            hidden_width, config.layer.layer_norm_eps, device=device
        )

    @classmethod
    def is_supported(cls: Type[Self], config: Dict[str, Any]) -> bool:
        return config.get("model_type") == "gpt_neox"

    @classmethod
    def state_dict_from_hf(
        cls: Type[Self], params: Mapping[str, Tensor]
    ) -> Mapping[str, Tensor]:
        return state_dict_from_hf(params, DECODER_HF_PARAM_KEY_TRANSFORMS)

    @classmethod
    def state_dict_to_hf(
        cls: Type[Self], params: Mapping[str, Tensor]
    ) -> Mapping[str, Tensor]:
        return state_dict_to_hf(params, DECODER_HF_PARAM_KEY_TRANSFORMS)

    @classmethod
    def config_from_hf(cls, hf_config: Mapping[str, Any]) -> GPTNeoXConfig:
        return _config_from_hf(hf_config)

    @classmethod
    def config_to_hf(cls, curated_config: GPTNeoXConfig) -> Mapping[str, Any]:
        return _config_to_hf(cls, curated_config)

    @classmethod
    def from_hf_config(
        cls: Type[Self],
        *,
        hf_config: Any,
        device: Optional[torch.device] = None,
    ) -> Self:
        config = cls.config_from_hf(hf_config)
        return cls(config, device=device)



from typing import TypeVar

from ..models.mpt import MPTCausalLM
from ..tokenizers.tokenizer import Tokenizer
from .default_generator import DefaultGenerator
from .hf_hub import FromHF


class MPTGenerator(DefaultGenerator, FromHF):
    """
    Generator for MPT model variants.
    """

    def __init__(self, tokenizer: Tokenizer, causal_lm: MPTCausalLM):
        """
        Construct an MPT generator.

        :param tokenizer:
            An MPT tokenizer.
        :param causal_lm:
            An MPT causal language model.
        """
        super().__init__(
            tokenizer,
            causal_lm,
        )



from typing import Optional

import torch
from torch import Tensor
from torch.nn import Module, Parameter


class RMSNorm(Module):
    """
    Root Mean Square (RMS) normalization (`Zhang et al., 2019`_).

    .. _Zhang et al., 2019: https://arxiv.org/abs/1910.07467
    """

    def __init__(
        self, width: int, *, eps: float, device: Optional[torch.device] = None
    ):
        """
        Construct a RMS normalization module.

        :param width:
            The (hidden) width of the representations that RMS
            normalization will be applied to.
        :param eps:
            Epsilon to avoid division by zero.
        :param device:
            Device on which the module is to be initialized.
        """
        super().__init__()
        self.eps = eps
        self.weight = Parameter(torch.ones((width,), device=device))

    def forward(self, input: Tensor) -> Tensor:
        """
        Apply RMS normalization to a tensor.

        :param input:
            The tensor to apply normalization to.
        :returns:
            Normalized tensor.
        """
        # Zhang & Sennrich, Equation 4. If we are in lower precision than
        # float32, then squaring and averaging can get way off. So for
        # normalization we want to use higher precision.
        rms = (
            input.to(torch.float32)
            .square()
            .mean(-1, keepdim=True)
            .add(self.eps)
            .rsqrt()
        )

        return (input * rms).to(input.dtype) * self.weight



import torch

from curated_transformers.generation.state import GeneratorState
from curated_transformers.generation.stop_conditions import (
    EndOfSequenceCondition,
    MaxGeneratedPiecesCondition,
)
from curated_transformers.layers.attention import AttentionMask


def test_end_of_sequence_condition():
    attention_mask = AttentionMask(torch.ones((2, 2), dtype=torch.bool))
    ids = torch.arange(0, 4).reshape(2, 2)
    state = GeneratorState(attention_mask=attention_mask, cache=None, prompt_ids=ids)
    state.generated_ids = torch.arange(4, 8).reshape(2, 2)

    completed_exclude = torch.zeros((2, 1), dtype=torch.bool)
    completed_include = torch.zeros((2, 1), dtype=torch.bool)
    EndOfSequenceCondition(4).update_completed(
        state=state,
        completed_exclude=completed_exclude,
        completed_include=completed_include,
    )
    assert not completed_exclude.any()
    assert not completed_include.any()

    EndOfSequenceCondition(6).update_completed(
        state=state,
        completed_exclude=completed_exclude,
        completed_include=completed_include,
    )
    assert not completed_exclude.any()
    assert not completed_include.any()

    EndOfSequenceCondition(5).update_completed(
        state=state,
        completed_exclude=completed_exclude,
        completed_include=completed_include,
    )
    assert completed_exclude.any()
    assert not completed_include.any()

    completed_exclude = torch.zeros((2, 1), dtype=torch.bool)
    EndOfSequenceCondition(7).update_completed(
        state=state,
        completed_exclude=completed_exclude,
        completed_include=completed_include,
    )
    assert completed_exclude.any()
    assert not completed_include.any()


def test_max_generated_pieces_condition():
    attention_mask = AttentionMask(torch.ones((2, 3), dtype=torch.bool))
    ids = torch.arange(0, 6).reshape(2, 3)
    state = GeneratorState(attention_mask=attention_mask, cache=None, prompt_ids=ids)
    state.generated_ids = torch.arange(4, 8).reshape(2, 2)

    completed_exclude = torch.zeros((2, 1), dtype=torch.bool)
    completed_include = torch.zeros((2, 1), dtype=torch.bool)
    MaxGeneratedPiecesCondition(3).update_completed(
        state=state,
        completed_exclude=completed_exclude,
        completed_include=completed_include,
    )
    assert not completed_exclude.any()
    assert not completed_include.any()

    MaxGeneratedPiecesCondition(2).update_completed(
        state=state,
        completed_exclude=completed_exclude,
        completed_include=completed_include,
    )
    assert not completed_exclude.any()
    assert completed_include.all()

    completed_include = torch.zeros((2, 1), dtype=torch.bool)
    MaxGeneratedPiecesCondition(1).update_completed(
        state=state,
        completed_exclude=completed_exclude,
        completed_include=completed_include,
    )
    assert not completed_exclude.any()
    assert completed_include.all()



import contextlib
import shutil
import tempfile
from pathlib import Path

import torch


@contextlib.contextmanager
def make_tempdir():
    d = Path(tempfile.mkdtemp())
    yield d
    shutil.rmtree(str(d))


# Wrapper around torch.testing.assert_close with custom tolerances.
def torch_assertclose(
    a: torch.Tensor, b: torch.Tensor, *, atol: float = 1e-05, rtol: float = 1e-05
):
    torch.testing.assert_close(
        a,
        b,
        atol=atol,
        rtol=rtol,
    )



from typing import Callable, Iterable, List

from ..tokenizer import PiecesWithIds
from ..util import remove_pieces_from_sequence
from .legacy_tokenizer import PostEncoder, PreDecoder


class FAIRSEQ_PIECE_IDS:
    FAIRSEQ_BOS = 0
    FAIRSEQ_EOS = 2
    FAIRSEQ_UNK = 3

    SPP_BOS = 1
    SPP_EOS = 2
    SPP_UNK = 0


class FairSeqPostEncoder(PostEncoder):
    """
    Performs fixups of SentencePiece piece identifiers for models that use
    the fairseq vocabulary.
    """

    def __init__(
        self,
        *,
        piece_updater: Callable[[int], int],
    ):
        """
        Construct a fairseq post-encoder.

        :param piece_updater:
            Function that tranforms a given SentencePiece piece
            identifier to a valid fairseq one.
        """
        self.piece_updater = piece_updater

    def __call__(self, pieces_with_ids: PiecesWithIds) -> PiecesWithIds:
        # We need to align the IDs to the original fairseq vocabulary.
        for piece_ids in pieces_with_ids.ids:
            for i in range(len(piece_ids)):
                piece_ids[i] = self.piece_updater(piece_ids[i])

        return pieces_with_ids


class FairSeqPreDecoder(PreDecoder):
    """
    Performs fixups of SentencePiece piece identifiers for models that use
    the fairseq vocabulary.
    """

    def __init__(
        self,
        *,
        bos_id: int,
        eos_id: int,
        piece_updater: Callable[[int], int],
    ):
        """
        Construct a fairseq pre-decoder.

        :param bos_id:
            The piece id used to mark the beginning of a sequence.
        :param eos_id:
            The piece id used to mark the end of a sequence.
        :param piece_updater:
            Function that tranforms a given fairseq piece identifier
            to the original SentencePiece one.
        """
        self.bos_id = bos_id
        self.eos_id = eos_id
        self.piece_updater = piece_updater

    def __call__(self, input: Iterable[Iterable[int]]) -> List[List[int]]:
        # Revert the fairseq alignment.
        input = (
            (self.piece_updater(piece_id) for piece_id in piece_ids)
            for piece_ids in input
        )

        return [
            list(remove_pieces_from_sequence(ids, (self.bos_id, self.eos_id)))
            for ids in input
        ]



import functools
from contextlib import contextmanager
from typing import List

import torch
from torch.nn import Module
from torch.utils.hooks import RemovableHandle

from .pytorch import ModuleIterator, apply_to_module


@contextmanager
def use_nvtx_ranges_for_forward_pass(module: Module):
    """
    Recursively applies `NVTX ranges`_ to the forward pass operation
    of the provided module. The ranges will be recorded during an `Nsight`_
    profiling session.

    :param module:
        Top-level module to which the ranges are applied recursively.

    .. _Nsight: https://developer.nvidia.com/nsight-systems
    .. _NVTX ranges: https://pytorch.org/docs/stable/cuda.html#nvidia-tools-extension-nvtx
    """

    hooks: List[RemovableHandle] = []

    def hook_forward(itr: ModuleIterator):
        range_name = f"{itr.name} : {type(itr.module).__name__}"

        def push(*args, _range_name: str, **kwargs):
            torch.cuda.nvtx.range_push(_range_name)

        def pop(*args, **kwargs):
            torch.cuda.nvtx.range_pop()

        forward_pre = itr.module.register_forward_pre_hook(
            functools.partial(push, _range_name=range_name)
        )
        forward_post = itr.module.register_forward_hook(pop)
        hooks.append(forward_pre)
        hooks.append(forward_post)

    try:
        apply_to_module(module, hook_forward)
        yield
    finally:
        for hook in hooks:
            hook.remove()



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, List

from torch import Tensor

if TYPE_CHECKING:
    from .state import GeneratorState


class StopCondition(ABC):
    """
    Base class for generation stop conditions.
    """

    @abstractmethod
    def update_completed(
        self,
        *,
        state: "GeneratorState",
        completed_exclude: Tensor,
        completed_include: Tensor
    ):
        """
        Update completed sequences according to the stop condition.

        :param completed_exclude:
            Output tensor marking which sequences are completed and
            for which the last generated piece **should not** be emitted.
        :param completed_include:
            Output tensor marking which sequences are completed and
            for which the last generated piece **should** be emitted.
        """
        ...


class CompoundStopCondition(List[StopCondition], StopCondition):
    """
    Sequentially apply multiple stop conditions.
    """

    def update_completed(
        self,
        *,
        state: "GeneratorState",
        completed_exclude: Tensor,
        completed_include: Tensor
    ):
        for condition in self:
            condition.update_completed(
                state=state,
                completed_exclude=completed_exclude,
                completed_include=completed_include,
            )


class EndOfSequenceCondition(StopCondition):
    """
    Stop when the end-of-sequence piece is predicted.
    """

    def __init__(self, eos_id: int):
        """
        Construct the stop condition.

        :param eos_id:
            End-of-sequence identifier that marks the end of a generated
            sequence.
        """
        self.eos_id = eos_id

    def update_completed(
        self,
        *,
        state: "GeneratorState",
        completed_exclude: Tensor,
        completed_include: Tensor
    ):
        completed_exclude ^= state.last_step_ids == self.eos_id


class MaxGeneratedPiecesCondition(StopCondition):
    """
    Stop after generating a maximum number of pieces.
    """

    def __init__(self, max_generated_pieces: int):
        """
        Construct the stop condition.

        :param max_generated_pieces:
            The maximum number of generated pieces. This condition is a noop
            for values less than 1.
        """
        super().__init__()
        self.max_generated_pieces = max_generated_pieces

    def update_completed(
        self,
        *,
        state: "GeneratorState",
        completed_exclude: Tensor,
        completed_include: Tensor
    ):
        if self.max_generated_pieces < 1:
            return False

        if state.generated_ids.size(-1) >= self.max_generated_pieces:
            completed_include ^= True



from dataclasses import dataclass

import torch

from ...layers.activations import Activation
from ..config import (
    TransformerAttentionLayerConfig,
    TransformerConfig,
    TransformerEmbeddingLayerConfig,
    TransformerFeedForwardLayerConfig,
    TransformerLayerConfig,
)


@dataclass
class ALBERTLayerConfig(TransformerLayerConfig):
    """
    ALBERT (`Lan et al., 2022`_) layer configuration.

    .. _Lan et al., 2022: https://arxiv.org/abs/1909.11942
    """

    n_layers_per_group: int
    n_hidden_groups: int

    def __init__(
        self, *args, n_layers_per_group: int = 1, n_hidden_groups: int = 1, **kwargs
    ):
        """
        :param n_layers_per_group:
            Number of layers per layer group.
        :param n_hidden_groups:
            Number of hidden groups.
        """
        super(ALBERTLayerConfig, self).__init__(*args, **kwargs)
        self.n_layers_per_group = n_layers_per_group
        self.n_hidden_groups = n_hidden_groups


@dataclass
class ALBERTConfig(TransformerConfig):
    """
    ALBERT (`Lan et al., 2022`_) model configuration.

    .. _Lan et al., 2022: https://arxiv.org/abs/1909.11942
    """

    layer: ALBERTLayerConfig

    def __init__(
        self,
        *,
        dtype: torch.dtype = torch.float32,
        embedding_width: int = 128,
        hidden_width: int = 768,
        n_layers_per_group: int = 1,
        intermediate_width: int = 3072,
        n_attention_heads: int = 12,
        n_hidden_layers: int = 12,
        n_hidden_groups: int = 1,
        attention_probs_dropout_prob: float = 0.0,
        hidden_dropout_prob: float = 0.0,
        activation: Activation = Activation.GELUNew,
        n_pieces: int = 30000,
        n_types: int = 2,
        n_positions: int = 512,
        model_max_length: int = 512,
        layer_norm_eps: float = 1e-12,
    ):
        """
        :param dtype:
            Data type to use for model parameters.
        :param embedding_width:
            Width of the embedding representations.
        :param hidden_width:
            Width of the transformer hidden layers.
        :param n_layers_per_group:
            Number of layers per layer group.
        :param intermediate_width:
            Width of the intermediate projection layer in the
            point-wise feed-forward layer.
        :param n_attention_heads:
            Number of self-attention heads.
        :param n_hidden_layers:
            Number of hidden layers.
        :param n_hidden_groups:
            Number of hidden groups.
        :param attention_probs_dropout_prob:
            Dropout probabilty of the self-attention layers.
        :param hidden_dropout_prob:
            Dropout probabilty of the point-wise feed-forward and
            embedding layers.
        :param activation:
            Activation used by the pointwise feed-forward layers.
        :param n_pieces:
            Size of main vocabulary.
        :param n_types:
            Size of token type vocabulary.
        :param n_positions:
            Maximum length of position embeddings.
        :param model_max_length:
            Maximum length of model inputs.
        :param layer_norm_eps:
            Epsilon for layer normalization.
        """
        self.embedding = TransformerEmbeddingLayerConfig(
            embedding_width=embedding_width,
            n_pieces=n_pieces,
            n_types=n_types,
            n_positions=n_positions,
            layer_norm_eps=layer_norm_eps,
            dropout_prob=hidden_dropout_prob,
        )
        self.layer = ALBERTLayerConfig(
            attention=TransformerAttentionLayerConfig(
                hidden_width=hidden_width,
                dropout_prob=attention_probs_dropout_prob,
                n_key_value_heads=n_attention_heads,
                n_query_heads=n_attention_heads,
                rotary_embeddings=None,
                use_alibi=False,
                use_bias=True,
                use_parallel_attention=False,
            ),
            feedforward=TransformerFeedForwardLayerConfig(
                hidden_width=hidden_width,
                intermediate_width=intermediate_width,
                activation=activation,
                use_bias=True,
                use_gate=False,
            ),
            n_hidden_layers=n_hidden_layers,
            layer_norm_eps=layer_norm_eps,
            dropout_prob=hidden_dropout_prob,
            n_layers_per_group=n_layers_per_group,
            n_hidden_groups=n_hidden_groups,
        )
        self.dtype = dtype
        self.model_max_length = model_max_length



from typing import Any, Dict, Mapping, Optional, Type, TypeVar

from curated_tokenizers import SentencePieceProcessor

from ...repository.file import RepositoryFile
from ..hf_hub import LegacyFromHF
from ._fairseq import FAIRSEQ_PIECE_IDS, FairSeqPostEncoder, FairSeqPreDecoder
from .legacy_tokenizer import AddBosEosPreEncoder
from .sentencepiece_tokenizer import SentencePieceTokenizer

# Only provided as typing.Self in Python 3.11+.
Self = TypeVar("Self", bound="CamemBERTTokenizer")


_CAMEMBERT_FAIRSEQ_OFFSET = 4


class CamemBERTPostEncoder(FairSeqPostEncoder):
    def __init__(
        self,
    ):
        """
        Construct a CamemBERT post-encoder.
        """
        super(CamemBERTPostEncoder, self).__init__(
            piece_updater=CamemBERTPostEncoder._sentencepiece_to_fairseq,
        )

    @staticmethod
    def _sentencepiece_to_fairseq(piece_id: int):
        if piece_id == FAIRSEQ_PIECE_IDS.SPP_UNK:
            return FAIRSEQ_PIECE_IDS.FAIRSEQ_UNK
        else:
            return piece_id + _CAMEMBERT_FAIRSEQ_OFFSET


class CamemBERTPreDecoder(FairSeqPreDecoder):
    def __init__(
        self,
        *,
        bos_id: int,
        eos_id: int,
    ):
        """
        Construct a CamemBERT pre-decoder.

        :param bos_id:
            The piece id used to mark the beginning of a sequence.
        :param eos_id:
            The piece id used to mark the end of a sequence.
        """
        self.bos_id = bos_id
        self.eos_id = eos_id
        super(CamemBERTPreDecoder, self).__init__(
            bos_id=bos_id,
            eos_id=eos_id,
            piece_updater=CamemBERTPreDecoder._fairseq_to_sentencepiece,
        )

    @staticmethod
    def _fairseq_to_sentencepiece(piece_id: int):
        if piece_id == FAIRSEQ_PIECE_IDS.FAIRSEQ_UNK:
            return FAIRSEQ_PIECE_IDS.SPP_UNK
        else:
            return piece_id - _CAMEMBERT_FAIRSEQ_OFFSET


class CamemBERTTokenizer(SentencePieceTokenizer, LegacyFromHF):
    """
    Legacy tokenizer for CamemBERT (`Martin et al., 2020`_) models.

    .. _Martin et al., 2020: https://arxiv.org/abs/1911.03894
    """

    vocab_files: Dict[str, str] = {"model": "sentencepiece.bpe.model"}

    def __init__(
        self,
        *,
        processor: SentencePieceProcessor,
        bos_piece: str = "<s>",
        eos_piece: str = "</s>",
    ):
        """
        Construct a CamemBERT tokenizer from a ``curated-tokenizers`` SentencePiece processor.

        :param processor:
            The processor to wrap.
        :param bos_piece:
            The piece to use to mark the beginning of a sequence.
        :param eos_piece:
            The piece to use to mark the end of a sequence.
        """
        super().__init__(processor=processor)

        self.processor = processor

        bos_id = _get_piece_id_or_fail(processor, bos_piece)
        eos_id = _get_piece_id_or_fail(processor, eos_piece)

        self.pre_encoder = AddBosEosPreEncoder(bos_piece=bos_piece, eos_piece=eos_piece)

        self.post_encoder = CamemBERTPostEncoder()

        self.pre_decoder = CamemBERTPreDecoder(
            bos_id=bos_id,
            eos_id=eos_id,
        )

    @classmethod
    def from_files(
        cls: Type[Self],
        *,
        model_file: RepositoryFile,
        bos_piece: str = "<s>",
        eos_piece: str = "</s>",
    ) -> Self:
        """
        Construct a tokenizer from vocabulary and merge files.

        :param model_file:
            The SentencePiece model file.
        :param bos_piece:
            The piece to use to mark the beginning of a sequence.
        :param eos_piece:
            The piece to use to mark the end of a sequence.
        """
        with model_file.open() as f:
            processor = SentencePieceProcessor.from_file(f)
        return cls(
            processor=processor,
            bos_piece=bos_piece,
            eos_piece=eos_piece,
        )

    @classmethod
    def _load_from_vocab_files(
        cls: Type[Self],
        *,
        vocab_files: Mapping[str, RepositoryFile],
        tokenizer_config: Optional[Dict[str, Any]],
    ) -> Self:
        return cls.from_files(model_file=vocab_files["model"])


def _get_piece_id_or_fail(processor: SentencePieceProcessor, piece: str) -> int:
    piece_id = processor.piece_to_id(piece)
    if piece_id == processor.unk_id():
        raise ValueError(
            f"CamemBERT piece encoder vocabulary doesn't contain '{piece}' piece"
        )
    return piece_id



from typing import List, TypeVar

from ..models.falcon import FalconCausalLM
from ..tokenizers.chunks import InputChunks, TextChunk
from ..tokenizers.tokenizer import Tokenizer
from .default_generator import DefaultGenerator
from .hf_hub import FromHF

# Only provided as typing.Self in Python 3.11+.
Self = TypeVar("Self", bound="FalconGenerator")


class FalconGenerator(DefaultGenerator, FromHF):
    """
    Generator for Falcon model variants.
    """

    def __init__(self, tokenizer: Tokenizer, causal_lm: FalconCausalLM):
        """
        Construct a Falcon generator.

        :param tokenizer:
            A Falcon tokenizer.
        :param causal_lm:
            A Falcon causal language model.

        """
        super().__init__(tokenizer, causal_lm)

    def preprocess_prompts(self, prompts: List[str]) -> List[InputChunks]:
        return [
            InputChunks(
                [
                    TextChunk(prompt.strip()),
                    # Add a newline, otherwise Falcon will generate an answer that
                    # starts with a newline.
                    TextChunk("\n"),
                ]
            )
            for prompt in prompts
        ]



import pytest
import torch

from curated_transformers.repository.file import LocalFile
from curated_transformers.tokenizers import PiecesWithIds
from curated_transformers.tokenizers.legacy.xlmr_tokenizer import XLMRTokenizer

from ...compat import has_hf_transformers
from ...utils import torch_assertclose
from ..util import compare_tokenizer_outputs_with_hf_tokenizer


@pytest.fixture
def toy_tokenizer(test_dir):
    return XLMRTokenizer.from_files(
        model_file=LocalFile(path=test_dir / "toy.model"),
    )


@pytest.mark.skipif(not has_hf_transformers, reason="requires huggingface transformers")
def test_from_hf_hub_equals_hf_tokenizer(sample_texts):
    compare_tokenizer_outputs_with_hf_tokenizer(
        sample_texts, "xlm-roberta-base", XLMRTokenizer
    )


def test_xlmr_toy_tokenizer(toy_tokenizer, short_sample_texts):
    encoding = toy_tokenizer(short_sample_texts)
    _check_toy_tokenizer(encoding)

    decoded = toy_tokenizer.decode(encoding.ids)
    assert decoded == [
        "I saw a girl with a telescope.",
        "Today we will eat pok ⁇  bowl, lots of it!",
        "Tokens which are unknown in ⁇  most ⁇  latin ⁇  alphabet ⁇  vocabularies.",
    ]


def _check_toy_tokenizer(pieces):
    assert isinstance(pieces, PiecesWithIds)
    assert len(pieces.ids) == 3
    assert len(pieces.pieces) == 3

    assert pieces.ids == [
        [0, 9, 466, 11, 948, 42, 11, 171, 169, 111, 29, 21, 144, 5, 2],
        [
            0,
            484,
            547,
            113,
            172,
            568,
            63,
            21,
            46,
            3,
            85,
            116,
            28,
            4,
            149,
            227,
            7,
            14,
            26,
            147,
            2,
        ],
        [
            0,
            484,
            46,
            95,
            7,
            140,
            123,
            222,
            46,
            25,
            116,
            25,
            20,
            3,
            637,
            3,
            149,
            77,
            54,
            3,
            11,
            28,
            30,
            53,
            19,
            66,
            16,
            15,
            3,
            8,
            84,
            21,
            29,
            19,
            66,
            232,
            50,
            458,
            5,
            2,
        ],
    ]

    assert pieces.pieces == [
        [
            "<s>",
            "▁I",
            "▁saw",
            "▁a",
            "▁girl",
            "▁with",
            "▁a",
            "▁t",
            "el",
            "es",
            "c",
            "o",
            "pe",
            ".",
            "</s>",
        ],
        [
            "<s>",
            "▁To",
            "day",
            "▁we",
            "▁will",
            "▁eat",
            "▁p",
            "o",
            "k",
            "é",
            "▁b",
            "ow",
            "l",
            ",",
            "▁l",
            "ot",
            "s",
            "▁of",
            "▁it",
            "!",
            "</s>",
        ],
        [
            "<s>",
            "▁To",
            "k",
            "en",
            "s",
            "▁which",
            "▁are",
            "▁un",
            "k",
            "n",
            "ow",
            "n",
            "▁in",
            "ペ",
            "▁most",
            "で",
            "▁l",
            "at",
            "in",
            "が",
            "▁a",
            "l",
            "p",
            "h",
            "a",
            "b",
            "e",
            "t",
            "際",
            "▁",
            "v",
            "o",
            "c",
            "a",
            "b",
            "ul",
            "ar",
            "ies",
            ".",
            "</s>",
        ],
    ]

    torch_assertclose(
        pieces.padded_tensor(padding_id=1),
        torch.tensor(
            [
                [
                    0,
                    9,
                    466,
                    11,
                    948,
                    42,
                    11,
                    171,
                    169,
                    111,
                    29,
                    21,
                    144,
                    5,
                    2,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                ],
                [
                    0,
                    484,
                    547,
                    113,
                    172,
                    568,
                    63,
                    21,
                    46,
                    3,
                    85,
                    116,
                    28,
                    4,
                    149,
                    227,
                    7,
                    14,
                    26,
                    147,
                    2,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                ],
                [
                    0,
                    484,
                    46,
                    95,
                    7,
                    140,
                    123,
                    222,
                    46,
                    25,
                    116,
                    25,
                    20,
                    3,
                    637,
                    3,
                    149,
                    77,
                    54,
                    3,
                    11,
                    28,
                    30,
                    53,
                    19,
                    66,
                    16,
                    15,
                    3,
                    8,
                    84,
                    21,
                    29,
                    19,
                    66,
                    232,
                    50,
                    458,
                    5,
                    2,
                ],
            ],
            dtype=torch.int32,
        ),
    )

    torch_assertclose(
        pieces.attention_mask().bool_mask.squeeze(dim=(1, 2)),
        torch.tensor(
            [
                [
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                ],
                [
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                ],
                [
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                ],
            ]
        ),
    )



import pytest

from curated_transformers.repository import HfHubRepository

TEST_LINES = ["Line 1", "Line 2 🐍", "🤖" "\n\n", ""]


@pytest.mark.upload_tests
def test_hf_hub_upload_txt():
    repo = HfHubRepository("explosion-testing/hf-upload-test")

    write_file = repo.file("test.txt")
    str_to_write = "\n".join(TEST_LINES)
    with write_file.open(mode="w") as f:
        f.write(str_to_write)

    read_file = repo.file("test.txt")
    with read_file.open(mode="r", encoding="utf-8") as f:
        str_read = f.read()
        assert str_read == str_to_write


@pytest.mark.upload_tests
def test_hf_hub_upload_binary():
    repo = HfHubRepository("explosion-testing/hf-upload-test")

    write_file = repo.file("test.bin")
    bytes_to_write = b"\x0D\x0E\x0A\x0D\x0B\x0E\x0E\x0F"
    with write_file.open(mode="wb") as f:
        f.write(bytes_to_write)

    read_file = repo.file("test.bin")
    with read_file.open(mode="rb") as f:
        bytes_read = f.read()
        assert bytes_read == bytes_to_write


@pytest.mark.upload_tests
def test_hf_hub_transactions():
    repo = HfHubRepository("explosion-testing/hf-upload-test")

    with repo.transaction() as tx:
        with tx.open("inner/test.txt", mode="w", encoding="utf-8") as f:
            str_to_write = "\n".join(TEST_LINES)
            f.write(str_to_write)
        with tx.open("test.bin", mode="wb") as f:
            bytes_to_write = b"\x0D\x0E\x0A\x0D\x0B\x0E\x0E\x0F"
            f.write(bytes_to_write)

    with repo.file("inner/test.txt").open(mode="r", encoding="utf-8") as f:
        str_read = f.read()
        assert str_read == str_to_write
    with repo.file("test.bin").open(mode="rb") as f:
        bytes_read = f.read()
        assert bytes_read == bytes_to_write


def test_hf_hub_failures():
    repo = HfHubRepository("explosion-testing/hf-upload-test")
    write_file = repo.file("nonexistent.bin")

    assert write_file.path is None
    with pytest.raises(FileNotFoundError):
        with write_file.open("r") as f:
            pass

    with pytest.raises(OSError):
        with write_file.open("a") as f:
            pass

    with pytest.raises(OSError, match="mode 'a' is not supported"):
        with repo.transaction() as tx:
            f = tx.open("test.txt", mode="a", encoding="utf-8")

    with pytest.raises(OSError, match="mode 'rb' is not supported"):
        with repo.transaction() as tx:
            f = tx.open("test.bin", mode="rb")

    with pytest.raises(OSError, match="already been opened"):
        with repo.transaction() as tx:
            f = tx.open("test.bin", mode="w")
            f1 = tx.open("test.bin", mode="w")



from typing import Any, Dict, Optional, Type, cast

from fsspec import AbstractFileSystem

from ..repository.fsspec import FsspecArgs, FsspecRepository
from ..repository.hf_hub import HfHubRepository
from ..repository.repository import Repository, TokenizerRepository
from .hf_hub import FromHF
from .legacy.bert_tokenizer import BERTTokenizer
from .legacy.camembert_tokenizer import CamemBERTTokenizer
from .legacy.llama_tokenizer import LlamaTokenizer
from .legacy.roberta_tokenizer import RoBERTaTokenizer
from .legacy.xlmr_tokenizer import XLMRTokenizer
from .tokenizer import Tokenizer, TokenizerBase

HF_TOKENIZER_MAPPING: Dict[str, Type[FromHF]] = {
    "BertTokenizer": BERTTokenizer,
    "BertTokenizerFast": BERTTokenizer,
    "CamembertTokenizer": CamemBERTTokenizer,
    "CamembertTokenizerFast": CamemBERTTokenizer,
    "LlamaTokenizer": LlamaTokenizer,
    "LlamaTokenizerFast": LlamaTokenizer,
    "RobertaTokenizer": RoBERTaTokenizer,
    "RobertaTokenizerFast": RoBERTaTokenizer,
    "XLMRobertaTokenizer": XLMRTokenizer,
    "XLMRobertaTokenizerFast": XLMRTokenizer,
}

HF_MODEL_MAPPING: Dict[str, Type[FromHF]] = {
    "bert": BERTTokenizer,
    "camembert": CamemBERTTokenizer,
    "llama": LlamaTokenizer,
    "roberta": RoBERTaTokenizer,
    "xlm-roberta": XLMRTokenizer,
}


class AutoTokenizer:
    """
    Tokenizer loaded from the Hugging Face Model Hub.
    """

    # NOTE: We do not inherit from FromHF, because its from_hf_hub method
    #       requires that the return type is Self.

    @classmethod
    def from_hf_hub_to_cache(
        cls,
        *,
        name: str,
        revision: str = "main",
    ):
        """
        Download the tokenizer's serialized model, configuration and vocab files
        from Hugging Face Hub into the local Hugging Face cache directory.
        Subsequent loading of the tokenizer will read the files from disk. If the
        files are already cached, this is a no-op.

        :param name:
            Model name.
        :param revision:
            Model revision.
        """
        tokenizer_cls = _resolve_tokenizer_class(
            TokenizerRepository(HfHubRepository(name, revision=revision))
        )
        tokenizer_cls.from_hf_hub_to_cache(name=name, revision=revision)

    @classmethod
    def from_fsspec(
        cls,
        *,
        fs: AbstractFileSystem,
        model_path: str,
        fsspec_args: Optional[FsspecArgs] = None,
    ) -> TokenizerBase:
        """
        Construct a tokenizer and load its parameters from an fsspec filesystem.

        :param fs:
            Filesystem.
        :param model_path:
            The model path.
        :param fsspec_args:
            Implementation-specific keyword arguments to pass to fsspec
            filesystem operations.
        :returns:
            The tokenizer.
        """
        return cls.from_repo(FsspecRepository(fs, model_path, fsspec_args))

    @classmethod
    def from_repo(cls, repo: Repository) -> TokenizerBase:
        """
        Construct and load a tokenizer from a repository.

        :param repository:
            The repository to load from.
        :returns:
            Loaded tokenizer.
        """
        tokenizer_cls = _resolve_tokenizer_class(TokenizerRepository(repo))
        return cast(
            TokenizerBase,
            tokenizer_cls.from_repo(repo),
        )

    @classmethod
    def from_hf_hub(cls, *, name: str, revision: str = "main") -> TokenizerBase:
        """
        Infer a tokenizer type and load it from Hugging Face Hub.

        :param name:
            Model name.
        :param revision:
            Model revision.
        :returns:
            The tokenizer.
        """
        return cls.from_repo(HfHubRepository(name, revision=revision))


def _get_tokenizer_class_from_config(
    tokenizer_config: Dict[str, Any]
) -> Optional[Type[FromHF]]:
    """
    Infer the tokenizer class from the tokenizer configuration.

    :param tokenizer_config:
        The tokenizer configuration.
    :param revision:
        Model revision.
    :returns:
        Inferred class.
    """
    return HF_TOKENIZER_MAPPING.get(tokenizer_config.get("tokenizer_class", None), None)


def _resolve_tokenizer_class(
    repo: TokenizerRepository,
) -> Type[FromHF]:
    tokenizer_file = repo.tokenizer_json()
    if tokenizer_file.exists():
        return Tokenizer

    cls: Optional[Type[FromHF]] = None
    try:
        tokenizer_config = repo.tokenizer_config()
        cls = _get_tokenizer_class_from_config(tokenizer_config)
    except:
        pass

    if cls is None:
        try:
            model_type = repo.model_type()
            print(model_type)
            cls = HF_MODEL_MAPPING.get(model_type)
        except:
            pass

    if cls is None:
        raise ValueError(f"Cannot infer tokenizer for repo: {repo.pretty_path()}")

    return cls



from typing import TypeVar

from ..models.llama import LlamaCausalLM
from ..tokenizers.tokenizer import Tokenizer
from .default_generator import DefaultGenerator
from .hf_hub import FromHF

# Only provided as typing.Self in Python 3.11+.
Self = TypeVar("Self", bound="LlamaGenerator")


class LlamaGenerator(DefaultGenerator, FromHF):
    """
    Generator for Llama and Llama 2 model variants.
    """

    def __init__(self, tokenizer: Tokenizer, causal_lm: LlamaCausalLM):
        """
        Construct a Llama generator.

        :param tokenizer:
            A Llama tokenizer.
        :param causal_lm:
            A Llama causal language model.

        """
        super().__init__(
            tokenizer,
            causal_lm,
        )



import importlib.metadata

import catalogue
from catalogue import Registry

__version__: str
try:
    __version__ = importlib.metadata.version("curated-transformers")
except importlib.metadata.PackageNotFoundError:
    __version__ = "unknown"


class registry(object):
    """
    Registry for models. These registries are used by auto classes to
    discover the available models.
    """

    causal_lms: Registry = catalogue.create(
        "curated_transformers", "causal_lms", entry_points=True
    )
    decoders: Registry = catalogue.create(
        "curated_transformers", "decoders", entry_points=True
    )
    encoders: Registry = catalogue.create(
        "curated_transformers", "encoders", entry_points=True
    )



from typing import Optional

from torch.nn import Module

from ..util.serde.load import TensorToParameterConverterT
from .bnb import prepare_for_quantization as bnb_prepare_for_quantization
from .bnb.config import BitsAndBytesConfig
from .quantizable import Quantizable


def prepare_module_for_quantization(
    module: Module, config: BitsAndBytesConfig
) -> Optional[TensorToParameterConverterT]:
    """
    Prepares a module for quantiazation and returns an optional callback
    to generate quantized parameter tensors.

    :param module:
        Top-level module to quantize. Should implement ``Quantizable``.
    :param config:
        Configuration for the quantizer.
    :returns:
        An optional callable that converts a non-quantized tensor
        to a quantized parameter.
    """
    if not isinstance(module, Quantizable):
        raise ValueError(f"Module of type `{type(module)}` is not quantizable")
    qmodel: Quantizable = module
    non_quantizable_module_prefixes = qmodel.modules_to_not_quantize()

    return bnb_prepare_for_quantization(module, config, non_quantizable_module_prefixes)



from typing import Optional

import torch
from torch import Tensor
from torch.nn import Linear, Module

from .activations import Activation


class PointwiseFeedForward(Module):
    """
    Point-wise feed-forward layer (`Vaswani et al., 2017`_).

    This layer is applied pointwise, meaning that the same
    transformation is applied to each sequence element. This
    transformation is:

    .. math::
        g(xW_1 + b_1)W_2 + b_2

    :math:`W_1` and :math:`b_1` transform the input to an
    intermediate width, :math:`g` is a non-linear activation
    function and :math:`W_2` and :math:`b_2` transform the
    output of the activation back to the input width.

    Gated Linear Units (`Dauphin et al., 2016`_; `Shazeer, 2020`_) are also
    supported. Gating applies the following transformation:

    .. math::
        (g(xW_g + b_g) * (xW_1 + b_1))W_2 + b_2

    :math:`W_g` and :math:`b_g` are the affine transformation for the gate.

    .. _Vaswani et al., 2017: https://arxiv.org/abs/1706.03762
    .. _Dauphin et al., 2016: https://arxiv.org/abs/1612.08083
    .. _Shazeer, 2020: https://arxiv.org/abs/2002.05202
    """

    gate: Optional[Linear]

    def __init__(
        self,
        *,
        activation: Module,
        hidden_width: int,
        intermediate_width: int,
        use_bias: bool,
        use_gate: bool,
        device: Optional[torch.device] = None,
    ):
        """
        Construct a pointwise feed-forward layer module.

        :param activation:
            Activation used by the pointwise feed-forward layers. The hidden
            input shape must be the same as the output shape (as is typical
            for elementwise activations).
        :param hidden_width:
            The input and output width of the layer.
        :param intermediate_width:
            The width of the projection to which the non-linearity is applied.
        :param use_bias:
            Use biases for linear layers.
        :param use_gate:
            Use Gated Linear Units.
        :param device:
            Device on which the module is to be initialized.
        """
        super().__init__()

        self.intermediate = Linear(
            hidden_width, intermediate_width, bias=use_bias, device=device
        )
        if use_gate:
            self.gate = Linear(
                hidden_width, intermediate_width, bias=use_bias, device=device
            )
        else:
            self.gate = None
        self.output = Linear(
            intermediate_width, hidden_width, bias=use_bias, device=device
        )
        self.activation = activation

    def forward(self, input: Tensor) -> Tensor:
        """
        Apply the point-wise feed-forward layer to the input.

        :param input:
            Input.

            *Shape:* ``(batch_size, seq_len, width)``
        :returns:
            Layer output.

            *Shape:* ``(batch_size, seq_len, width)``
        """
        if self.gate is None:
            return self.output(self.activation(self.intermediate(input)))
        else:
            return self.output(
                self.activation(self.gate(input)) * self.intermediate(input)
            )



from contextlib import contextmanager
from typing import Callable, Dict, Iterable, Mapping, Optional, Set, Union

import torch
from torch.nn import Module, Parameter

from ...repository.file import RepositoryFile
from ..pytorch import ModuleIterator, apply_to_module
from .checkpoint import ModelCheckpointType

# Args: Parent module, module prefix, parameter name, tensor to convert, device.
# Returns the new paramater.
TensorToParameterConverterT = Callable[
    [Module, str, str, torch.Tensor, Optional[torch.device]], Parameter
]

# Args: State dict.
# Returns the converted state dict.
HFStateDictConverterT = Callable[
    [Mapping[str, torch.Tensor]], Mapping[str, torch.Tensor]
]


def load_model_from_checkpoints(
    model: Module,
    *,
    filepaths: Iterable[RepositoryFile],
    checkpoint_type: ModelCheckpointType,
    state_dict_converter: HFStateDictConverterT,
    tensor_to_param_converter: Optional[TensorToParameterConverterT] = None,
    device: Optional[torch.device] = None,
):
    """
    Load parameters from PyTorch checkpoints with minimal copies.

    :param model:
        PyTorch module into which the parameters are to be loaded.
    :param filepaths:
        Paths to PyTorch checkpoints.
    :param checkpoint_type:
        Type of checkpoint being loaded.
    :param state_dict_converter:
        Callback to convert Hugging Face state dicts to the
        ``curated-transformers`` format.
    :param tensor_to_param_converter:
        Callback to perform custom conversions of the loaded parameters.
        Useful for loading quantized weights.
    :param device:
        Device in which to place the loaded parameters.
    """

    state_dicts = checkpoint_type.loader(filepaths)
    # We need to cache the model's parameter keys before loading the state
    # dicts as the process could potentially change the structure of sub-modules,
    # e.g: when quantized layers rename their parameters.
    module_keys = set(model.state_dict().keys())
    seen_keys: Set[str] = set()

    for state_dict in state_dicts:
        converted = state_dict_converter(state_dict)
        if len(converted) == 0:
            continue
        seen_keys.update(converted.keys())

        # We have to walk the module tree for each state dict as there
        # are no guarantees on the ordering of the keys.
        _emplace_module_state_dict(
            model,
            converted,
            tensor_to_param_converter=tensor_to_param_converter,
            device=device,
        )

    # Make sure that we didn't miss any keys.
    missing_keys = module_keys.difference(seen_keys)
    if len(missing_keys) != 0:
        raise ValueError(f"Some parameters were not updated/replaced: {missing_keys}")


def default_tensor_to_parameter_converter(
    module: Module,
    module_prefix: str,
    parameter_name: str,
    tensor: torch.Tensor,
    device: Optional[torch.device] = None,
) -> Parameter:
    """
    Default tensor to parameter converter.

    :param module:
        Parent module of the parameter being converted/replaced.
    :param module_prefix:
        Prefix of the parent module.
    :param parameter_name:
        Name of the parameter being converted/replaced.
    :param tensor:
        Tensor to be converted.
    :param device:
        Device to which the converted parameter is moved.
    :returns:
        Converted parameter.
    """
    old_param = module._parameters[parameter_name]
    assert old_param is not None
    _validate_replacement(old_param, tensor, module_prefix)
    return Parameter(tensor.to(device=device), requires_grad=old_param.requires_grad)  # type: ignore


def _emplace_module_state_dict(
    module: Module,
    state_dict: Mapping[str, torch.Tensor],
    *,
    tensor_to_param_converter: Optional[TensorToParameterConverterT] = None,
    device: Optional[torch.device] = None,
):
    if tensor_to_param_converter is None:
        tensor_to_param_converter = default_tensor_to_parameter_converter

    def apply(itr: ModuleIterator):
        prefix_with_dot = f"{itr.prefix}."
        candidate_tensors = {
            k: v for k, v in state_dict.items() if k.startswith(prefix_with_dot)
        }
        if len(candidate_tensors) == 0:
            return

        local_params_and_buffers: Dict[
            str, Union[Optional[Parameter], Optional[torch.Tensor]]
        ] = dict(itr.module._parameters.items())
        for name, buf in itr.module._buffers.items():
            if name in local_params_and_buffers:
                raise KeyError(
                    f"Key `{name}` used in both learnable parameters and buffers in module `{itr.prefix}`"
                )
            elif name not in itr.module._non_persistent_buffers_set:
                local_params_and_buffers[name] = buf

        for name, param in local_params_and_buffers.items():
            key = f"{prefix_with_dot}{name}"
            if key not in candidate_tensors:
                continue
            elif param is None:
                raise ValueError(
                    f"Key `{name}` found in state dict but no data in module `{itr.prefix}`"
                )
            replacement = candidate_tensors[key]
            assert tensor_to_param_converter is not None
            _emplace_module_tensor(
                module=itr.module,
                module_prefix=itr.prefix,
                tensor_name=name,
                replacement_tensor=replacement,
                tensor_to_param_converter=tensor_to_param_converter,
                device=device,
            )

    apply_to_module(module, apply)


def _emplace_module_tensor(
    module: Module,
    module_prefix: str,
    tensor_name: str,
    replacement_tensor: torch.Tensor,
    tensor_to_param_converter: TensorToParameterConverterT,
    device: Optional[torch.device] = None,
):
    """
    Replace a module's parameter or (persistent) buffer with the
    passed tensor and moves it to the given device.

    This is a zero-copy operation (excluding D2H/H2D transfers) where
    the input tensor is directly associated with the module. Unexpected
    behaviour can occur if the same tensor is associated with multiple modules.
    """
    is_parameter = tensor_name in module._parameters
    is_buffer = tensor_name in module._buffers
    assert is_parameter ^ is_buffer

    if is_parameter:
        new_param = tensor_to_param_converter(
            module, module_prefix, tensor_name, replacement_tensor, device
        )
        module._parameters[tensor_name] = new_param
    else:
        old_buffer = module._buffers[tensor_name]
        assert old_buffer is not None
        _validate_replacement(
            old_buffer, replacement_tensor, f"{module_prefix}.{tensor_name}"
        )
        module._buffers[tensor_name] = replacement_tensor


def _validate_replacement(
    replaced: Union[Parameter, torch.Tensor],
    replacement: torch.Tensor,
    name: str,
):
    if replaced.shape != replacement.shape:
        raise ValueError(
            f"Expected size of replacement for `{name}` to be {replaced.shape}, but got {replacement.shape}"
        )
    elif replaced.dtype != replacement.dtype:
        raise ValueError(
            f"Expected dtype of replacement for `{name}` to be {replaced.dtype}, but got {replacement.dtype}"
        )



from abc import ABC, abstractmethod
from typing import List

from .config import GeneratorConfig


class GeneratorWrapper(ABC):
    """
    Model-specific wrapper for :class:`curated_transformers.generation.Generator`.
    """

    def __call__(self, prompts: List[str], config: GeneratorConfig) -> List[str]:
        """
        Alias for :meth:`.generate`.
        """

        return self.generate(prompts, config)

    @abstractmethod
    def generate(self, prompts: List[str], config: GeneratorConfig) -> List[str]:
        """
        Generate text using the given prompts.

        :param prompts:
            Prompts to generate from.
        :param config:
            Generator configuraton.
        :returns:
            Strings generated for the prompts.
        """
        ...



from typing import TYPE_CHECKING

from ..util.serde.checkpoint import ModelCheckpointType

HF_MODEL_CONFIG = "config.json"
HF_MODEL_CHECKPOINT = "pytorch_model.bin"
HF_MODEL_CHECKPOINT_SAFETENSORS = "model.safetensors"
HF_MODEL_SHARDED_CHECKPOINT_INDEX = "pytorch_model.bin.index.json"
HF_MODEL_SHARDED_CHECKPOINT_INDEX_SAFETENSORS = "model.safetensors.index.json"
HF_MODEL_SHARDED_CHECKPOINT_INDEX_WEIGHTS_KEY = "weight_map"
HF_TOKENIZER_CONFIG = "tokenizer_config.json"
SPECIAL_TOKENS_MAP = "special_tokens_map.json"
TOKENIZER_JSON = "tokenizer.json"


PRIMARY_CHECKPOINT_FILENAMES = {
    ModelCheckpointType.PYTORCH_STATE_DICT: HF_MODEL_CHECKPOINT,
    ModelCheckpointType.SAFE_TENSORS: HF_MODEL_CHECKPOINT_SAFETENSORS,
}
SHARDED_CHECKPOINT_INDEX_FILENAMES = {
    ModelCheckpointType.PYTORCH_STATE_DICT: HF_MODEL_SHARDED_CHECKPOINT_INDEX,
    ModelCheckpointType.SAFE_TENSORS: HF_MODEL_SHARDED_CHECKPOINT_INDEX_SAFETENSORS,
}
# Same for both checkpoint types.
SHARDED_CHECKPOINT_INDEX_WEIGHTS_KEY = HF_MODEL_SHARDED_CHECKPOINT_INDEX_WEIGHTS_KEY



from setuptools import setup, find_packages


def setup_package():
    setup(
        name="curated-transformers",
        packages=find_packages(),
    )


if __name__ == "__main__":
    setup_package()



import json
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Type, TypeVar, Union, cast

import torch
from fsspec.implementations.local import LocalFileSystem
from huggingface_hub import Repository
from tokenizers import Tokenizer as HFTokenizer
from torch import Tensor

from ..layers.attention import AttentionMask
from ..repository.file import RepositoryFile
from ..repository.fsspec import FsspecRepository
from ..repository.hf_hub import HfHubRepository
from ..repository.repository import Repository, TokenizerRepository
from ._hf_compat import clean_up_decoded_string_like_hf
from .chunks import InputChunks, MergedSpecialPieceChunk
from .hf_hub import FromHF

# Only provided as typing.Self in Python 3.11+.
Self = TypeVar("Self", bound="Tokenizer")


@dataclass
class PiecesWithIds:
    """
    Encoded output of tokenizers.

    :param ids:
        Piece identifiers of each input sequence.
    :param pieces:
        Piece strings of each input sequence.
    """

    ids: List[List[int]]
    pieces: List[List[str]]

    def attention_mask(
        self, *, pad_left: bool = False, device: Optional[torch.device] = None
    ) -> AttentionMask:
        """
        Generate the attention masks. The mask is equivalent to:
        ``ids.padded_tensor(padding_id) != padding_id``

        :param pad_left:
            By default sequences shorter than the longest sequence are
            right-padded. Use left-padding when set to ``True``.
        :param device:
            Device on which the attention mask is created.
        :returns:
            The attention mask.

            *Shape:* ``(batch_size, max_seq_len)``
        """
        n_seqs = len(self.ids)
        max_len = max(len(seq_ids) for seq_ids in self.ids)
        mask = torch.full((n_seqs, max_len), False, device=device)
        for idx, seq_ids in enumerate(self.ids):
            if pad_left:
                mask[idx, -len(seq_ids) :] = True
            else:
                mask[idx, : len(seq_ids)] = True
        return AttentionMask(mask)

    def padded_tensor(
        self,
        *,
        padding_id: int = 0,
        pad_left: bool = False,
        device: Optional[torch.device] = None,
    ) -> Tensor:
        """
        Generate a padded tensor of the piece identifiers.

        :param padding_id:
            Piece identifier of the padding piece. The actual identifier
            generally doesn't matter when an attention mask is used (and
            as long as it is a valid vocabulary index).
        :param pad_left:
            By default sequences shorter than the longest sequence are
            right-padded. Use left-padding when set to ``True``.
        :param device:
            Device on which the padded tensor is created.
        :returns:
            The padded piece ids.

            *Shape:* ``(batch_size, max_seq_len)``
        """
        n_seqs = len(self.ids)
        max_len = max(len(seq_ids) for seq_ids in self.ids)
        padded = torch.full(
            (n_seqs, max_len), padding_id, dtype=torch.int32, device=device
        )
        for idx, seq_ids in enumerate(self.ids):
            if pad_left:
                padded[idx, -len(seq_ids) :] = torch.tensor(seq_ids)
            else:
                padded[idx, : len(seq_ids)] = torch.tensor(seq_ids)
        return padded


class TokenizerBase(ABC):
    """
    Base class for all tokenizers.
    """

    def __call__(
        self, input: Union[Iterable[InputChunks], Iterable[str]]
    ) -> PiecesWithIds:
        """
        Split one or more texts into pieces.

        :param input:
            Sequences to tokenize. If the sequences are strings, they are
            automatically converted to chunks.
        :returns:
            Pieces in each sequence.
        """
        return self.encode(input)

    @abstractmethod
    def decode(
        self, input: Iterable[Iterable[int]], *, skip_special_pieces: bool = True
    ) -> List[str]:
        """
        Reconstruct string sequences from piece identifiers.

        :param input:
            The piece identifiers to reconstruct the strings from.
        :param skip_special_pieces:
            Skip special pieces during decoding.
        :returns:
            The decoded strings.
        """
        ...

    @abstractmethod
    def encode(
        self, input: Union[Iterable[InputChunks], Iterable[str]]
    ) -> PiecesWithIds:
        """
        Split one or more texts into pieces.

        :param input:
            Sequences to tokenize. If the sequences are strings, they are
            automatically converted to chunks.
        :returns:
            Pieces in each sequence.
        """
        ...

    @abstractmethod
    def piece_to_id(self, piece: str) -> Optional[int]:
        """
        Get the ID for a single piece.

        :param piece:
            The piece to look up the identifier for.
        :returns:
            The piece identifier or ``None`` when the piece is unknown.
        """
        ...

    @property
    @abstractmethod
    def eos_piece(self) -> Optional[str]:
        """
        Get the end-of-sequence piece.

        :returns:
            The end-of-sequence piece or ``None`` when this piece is not
            defined.
        """
        ...


class Tokenizer(TokenizerBase, FromHF):
    """
    Wraps the tokenizers from the ``tokenizers`` package. It supports a
    wide range of piece tokenizers, including word piece, byte pair encoding, and
    sentencepiece unigram tokenizers. This is the tokenizer that should be used
    in the majority of cases. The other tokenizers in the ``curated-transformers``
    package should only be used when you have a legacy tokenizer that is not in
    Hugging Face ``tokenizer.json`` format.
    """

    def __init__(
        self,
        *,
        tokenizer: HFTokenizer,
        config: Optional[Dict[str, Any]],
        special_tokens_map: Optional[Dict[str, Any]],
    ):
        """
        Construct a tokenizer.

        :param tokenizer:
            The ``tokenizers`` tokenizer to use.
        :param config:
            Additional tokenizer configuration.
        :param special_tokens_map:
            Map of special tokens.
        """
        self.tokenizer = tokenizer
        self._eos_piece = self._get_special_piece(
            piece_name="eos_token",
            tokenizer_config=config,
            special_tokens_map=special_tokens_map,
        )

    def _get_special_piece(
        self,
        *,
        piece_name: str,
        tokenizer_config: Optional[Dict[str, Any]],
        special_tokens_map: Optional[Dict[str, Any]],
    ) -> Optional[str]:
        """
        Attempt to get the special piece identifier from the special tokens
        mapping or the tokenizer configuration.

        :param piece_name:
            The name of the special piece to look up.
        :param tokenizer_config:
            The tokenizer configuration.
        :param special_piece_map:
            Special pieces mapping.
        :returns:
            The special piece, if defined in the special tokens mapping or
            the tokenizer configuration.
        """
        piece = None
        if special_tokens_map is not None:
            piece = get_special_piece(special_tokens_map, piece_name)
        if piece is None and tokenizer_config is not None:
            piece = get_special_piece(tokenizer_config, piece_name)
        return piece

    def decode(
        self,
        input: Iterable[Iterable[int]],
        *,
        skip_special_pieces: bool = True,
    ) -> List[str]:
        decoded = self.tokenizer.decode_batch(
            input, skip_special_tokens=skip_special_pieces
        )

        decoded = [clean_up_decoded_string_like_hf(string) for string in decoded]

        return decoded

    def encode(
        self, input: Union[Iterable[InputChunks], Iterable[str]]
    ) -> PiecesWithIds:
        input = cast(Union[List[InputChunks], List[str]], list(input))
        if isinstance(input[0], str):
            input = cast(List[str], input)
            return self._encode_strings(input)
        else:
            input = cast(List[InputChunks], input)
            return self._encode_chunks(input)

    def _encode_strings(self, input: Iterable[str]) -> PiecesWithIds:
        non_str = {type(seq).__name__ for seq in input if not isinstance(seq, str)}
        if non_str:
            raise ValueError(f"Non-string inputs: {', '.join(list(sorted(non_str)))}")
        encodings = self.tokenizer.encode_batch(input)
        ids = [encoding.ids for encoding in encodings]
        pieces = [encoding.tokens for encoding in encodings]
        return PiecesWithIds(ids=ids, pieces=pieces)

    def _encode_chunks(self, input: Iterable[InputChunks]) -> PiecesWithIds:
        non_str = {
            type(seq).__name__ for seq in input if not isinstance(seq, InputChunks)
        }
        if non_str:
            raise ValueError(f"Non-chunk inputs: {', '.join(list(sorted(non_str)))}")

        merged_chunks = [seq_chunks.merge_text_chunks() for seq_chunks in input]

        ids = []
        pieces = []

        for seq in merged_chunks:
            seq_ids = []
            seq_pieces = []

            for chunk in seq:
                if isinstance(chunk, MergedSpecialPieceChunk):
                    piece_id = self.tokenizer.token_to_id(chunk.piece)
                    seq_ids.append(piece_id)
                    seq_pieces.append(chunk.piece)
                else:
                    encoding = self.tokenizer.encode(chunk.text)
                    seq_ids.extend(encoding.ids)
                    seq_pieces.extend(encoding.tokens)

            ids.append(seq_ids)
            pieces.append(seq_pieces)

        return PiecesWithIds(ids=ids, pieces=pieces)

    @property
    def eos_piece(self) -> Optional[str]:
        return self._eos_piece

    @classmethod
    def from_dir(cls: Type[Self], path: Path) -> Self:
        """
        Load the tokenizer from a directory with a ``tokenizer.json`` file.

        :param path:
            Path to the tokenizer directory.
        """
        return cls.from_repo(FsspecRepository(LocalFileSystem(), str(path)))

    @classmethod
    def from_hf_hub_to_cache(
        cls: Type[Self],
        *,
        name: str,
        revision: str = "main",
    ):
        repo = TokenizerRepository(HfHubRepository(name, revision=revision))
        repo.tokenizer_json()

        try:
            _ = repo.tokenizer_config()
        except:
            pass
        try:
            _ = repo.special_tokens_map()
        except:
            pass

    @classmethod
    def from_repo(cls: Type[Self], repo: Repository) -> Self:
        def initialize_hf_tokenizer(tokenizer_file: RepositoryFile) -> HFTokenizer:
            with tokenizer_file.open(mode="rb") as f:
                # We have to open the file once before checking the local
                # file's existence (since the repo files are loaded lazily).
                if tokenizer_file.path is not None:
                    return HFTokenizer.from_file(tokenizer_file.path)
                else:
                    return HFTokenizer.from_buffer(f.read())

        repo = TokenizerRepository(repo)
        tokenizer_file = repo.tokenizer_json()
        hf_tokenizer = initialize_hf_tokenizer(tokenizer_file)

        try:
            config = repo.tokenizer_config()
        except OSError:
            config = None
        try:
            special_tokens_map = repo.special_tokens_map()
        except OSError:
            special_tokens_map = None
        return cls(
            tokenizer=hf_tokenizer,
            config=config,
            special_tokens_map=special_tokens_map,
        )

    @classmethod
    def from_json(
        cls: Type[Self],
        tokenizer_json: str,
        config_json: Optional[str] = None,
        special_tokens_map_json: Optional[str] = None,
    ) -> Self:
        """
        Load the tokenizer from serialized JSON strings.

        :param tokenizer_json:
            The JSON string of the serialized tokenizer.
        :param config_json:
            The JSON string of the tokenizer config.
        :param special_tokens_map_json:
            The JSON string of the special tokens map.
        """
        hf_tokenizer = HFTokenizer.from_str(tokenizer_json)
        config = json.loads(config_json) if config_json is not None else None
        special_tokens_map = (
            json.loads(special_tokens_map_json)
            if special_tokens_map_json is not None
            else None
        )
        return cls(
            tokenizer=hf_tokenizer,
            config=config,
            special_tokens_map=special_tokens_map,
        )

    def piece_to_id(self, piece: str) -> Optional[int]:
        return self.tokenizer.token_to_id(piece)


def get_special_piece(
    special_tokens_map: Dict[str, Any], piece_name: str
) -> Optional[str]:
    """
    Get a special piece from the special tokens map or the tokenizer
    configuration.

    :param special_tokens_map:
        The special tokens map.
    :param piece_name:
        The piece to look up.
    :returns:
        The piece or ``None`` if this particular piece was not defined.
    """
    piece = special_tokens_map.get(piece_name)
    if isinstance(piece, dict):
        piece = piece.get("content")
    return piece



from typing import Optional, Tuple

import torch
from torch import Tensor
from torch.nn import Module

from ...layers.attention import (
    AttentionHeads,
    AttentionLinearBiases,
    AttentionMask,
    KeyValueCache,
    QkvMode,
    QkvSplitGroupedByKVHeads,
    ScaledDotProductAttention,
    SelfAttention,
)
from ...layers.embeddings import QueryKeyRotaryEmbeddings
from ...layers.feedforward import PointwiseFeedForward
from ..config import TransformerLayerConfig


class OldFalconDecoderLayer(Module):
    """
    Falcon (`Penedo et al., 2019`_) layer using the old decoder architecture.

    .. _Penedo et al., 2019: https://arxiv.org/abs/2306.01116
    """

    def __init__(
        self,
        layer_config: TransformerLayerConfig,
        *,
        device: Optional[torch.device] = None,
    ):
        super().__init__()

        attention_config = layer_config.attention
        if attention_config.rotary_embeddings is None:
            raise ValueError(
                "Falcon attention config does not contain rotary embedding parameters"
            )

        self.use_parallel_attention = attention_config.use_parallel_attention

        hidden_width = layer_config.feedforward.hidden_width
        n_attention_heads = attention_config.n_query_heads
        attention_biases = (
            AttentionLinearBiases(
                n_attention_heads=attention_config.n_query_heads,
                is_causal=True,
                is_inverted=True,
            )
            if attention_config.use_alibi
            else None
        )
        # Rotary embeddings are disabled when using ALiBi.
        rotary_embeds = (
            QueryKeyRotaryEmbeddings(
                fraction=attention_config.rotary_embeddings.rotary_fraction,
                base=attention_config.rotary_embeddings.rotary_base,
                head_width=hidden_width // n_attention_heads,
            )
            if not attention_config.use_alibi
            else None
        )
        self.mha = SelfAttention(
            attention_scorer=ScaledDotProductAttention(
                dropout_prob=attention_config.dropout_prob,
                linear_biases=attention_biases,
            ),
            hidden_width=hidden_width,
            attention_heads=AttentionHeads.key_value_broadcast(
                n_query_heads=attention_config.n_query_heads,
                n_key_value_heads=attention_config.n_key_value_heads,
                qkv_split=QkvSplitGroupedByKVHeads(),
            ),
            rotary_embeds=rotary_embeds,
            qkv_mode=(
                QkvMode.MERGED_SPLIT_AFTER
                if attention_config.n_key_value_heads == 1
                else QkvMode.MERGED_SPLIT_BEFORE
            ),
            use_bias=attention_config.use_bias,
            device=device,
        )
        self.attn_output_dropout = torch.nn.Dropout(p=layer_config.dropout_prob)
        self.attn_layer_norm = torch.nn.LayerNorm(
            hidden_width, eps=layer_config.layer_norm_eps, device=device
        )

        if not self.use_parallel_attention:
            self.ffn_layer_norm = torch.nn.LayerNorm(
                hidden_width,
                eps=layer_config.layer_norm_eps,
                device=device,
            )

        self.ffn = PointwiseFeedForward(
            activation=layer_config.feedforward.activation.module(),
            hidden_width=hidden_width,
            intermediate_width=layer_config.feedforward.intermediate_width,
            use_bias=layer_config.feedforward.use_bias,
            use_gate=layer_config.feedforward.use_gate,
            device=device,
        )
        self.ffn_output_dropout = torch.nn.Dropout(p=layer_config.dropout_prob)

    def forward(
        self,
        input: Tensor,
        attention_mask: AttentionMask,
        *,
        cache: Optional[KeyValueCache] = None,
        positions: Optional[Tensor] = None,
        store_cache: bool = False,
    ) -> Tuple[Tensor, Optional[KeyValueCache]]:
        """
        Apply the Falcon layer to the given piece hidden representations.

        :param input:
            Hidden representations to apply the layer to.

            *Shape:* ``(batch_size, seq_len, width)``
        :param attention_mask:
            Attention mask. Sequence elements for which the
            corresponding mask element is set to ``False`` are ignored
            during attention calculation.
        :param cache:
            Key/value cache to avoid recomputing key/value representations
            for tokens that were previously seen.
        :param positions:
            Input positions. Positions are needed to look up rotary embeddings.
            Normally, these positions are calculated automatically. But if the
            positions deviate for some reason, they can be provided through this
            argument.
        :param store_cache:
            Whether to cache the key/value representations for future reuse.
        :returns:
            Layer output and the key/value cache.
        """
        residual = input

        attn_layer_norm = self.attn_layer_norm(input)

        attn_out, cache = self.mha(
            attn_layer_norm,
            attention_mask,
            cache=cache,
            store_cache=store_cache,
            positions=positions,
            use_causal_mask=True,
        )

        if self.use_parallel_attention:
            ffn_layer_norm = attn_layer_norm
        else:
            residual = residual + self.attn_output_dropout(attn_out)
            ffn_layer_norm = self.ffn_layer_norm(residual)

        ffn_out = self.ffn(ffn_layer_norm)

        if self.use_parallel_attention:
            ffn_out += attn_out

        return residual + self.ffn_output_dropout(ffn_out), cache



from abc import abstractmethod
from typing import Generic, List, Optional, TypeVar

from torch import Tensor
from torch.nn import Module

from ..layers.attention import AttentionMask
from .config import TransformerConfig
from .output import CacheT, CausalLMOutputWithCache, ModelOutput, ModelOutputWithCache

ConfigT = TypeVar("ConfigT", bound=TransformerConfig)


class TransformerModule(Generic[ConfigT], Module):
    """
    Base class for transformer modules.
    """

    _config: ConfigT

    def __init__(self, config: ConfigT):
        super().__init__()

        self._config = config

    @property
    def config(self) -> ConfigT:
        """
        Returns the model's configuration.
        """
        return self._config


class CausalLMModule(Generic[ConfigT, CacheT], TransformerModule[ConfigT]):
    """
    Base class for causal language model modules.
    """

    def __init__(self, config: ConfigT):
        super().__init__(config)

    @abstractmethod
    def forward(
        self,
        piece_ids: Tensor,
        attention_mask: AttentionMask,
        *,
        cache: Optional[List[CacheT]] = None,
        positions: Optional[Tensor] = None,
        store_cache: bool = False,
    ) -> CausalLMOutputWithCache[CacheT]:
        """
        Apply the causal language model to the given piece identifiers.

        :param piece_ids:
            Piece identifiers to apply the decoder to.

            *Shape:* ``(batch_size, seq_len)``
        :param attention_mask:
            Attention mask. Sequence elements for which the
            corresponding mask element is set to ``False`` are ignored
            during attention calculation.
        :param cache:
            Key/value cache to avoid recomputing key/value representations
            for tokens that were previously seen.
        :param positions:
            Input positions. Positions are needed to look up rotary embeddings.
            Normally, these positions are calculated automatically. But if the
            positions deviate for some reason, they can be provided through this argument.

            *Shape:* ``(batch_size, seq_len)``
        :param store_cache:
            Whether to cache the key/value representations for future reuse.
        :returns:
            Causal language model output with key/value cache.
        """
        raise NotImplementedError


class DecoderModule(Generic[ConfigT, CacheT], TransformerModule[ConfigT]):
    """
    Base class for decoder modules.
    """

    def __init__(self, config: ConfigT):
        super().__init__(config)

    @abstractmethod
    def forward(
        self,
        piece_ids: Tensor,
        attention_mask: AttentionMask,
        *,
        cache: Optional[List[CacheT]] = None,
        positions: Optional[Tensor] = None,
        store_cache: bool = False,
    ) -> ModelOutputWithCache[CacheT]:
        """
        Apply the decoder to the given piece identifiers.

        :param piece_ids:
            Piece identifiers to apply the decoder to.

            *Shape:* ``(batch_size, seq_len)``
        :param attention_mask:
            Attention mask. Sequence elements for which the
            corresponding mask element is set to ``False`` are ignored
            during attention calculation.
        :param cache:
            Key/value cache to avoid recomputing key/value representations
            for tokens that were previously seen.
        :param positions:
            Input positions. Positions are needed to look up position embeddings.
            Normally, these positions are calculated automatically. But if the
            positions deviate for some reason, they can be provided through this argument.

            *Shape:* ``(batch_size, seq_len)``
        :param store_cache:
            Whether to cache the key/value representations for future reuse.
        :returns:
            Decoder output with key/value cache.
        """
        raise NotImplementedError


class EncoderModule(Generic[ConfigT], TransformerModule[ConfigT]):
    """
    Base class for encoder modules.
    """

    def __init__(self, config: ConfigT):
        super().__init__(config)

    @abstractmethod
    def forward(
        self,
        piece_ids: Tensor,
        attention_mask: AttentionMask,
        *,
        positions: Optional[Tensor] = None,
        type_ids: Optional[Tensor] = None,
    ) -> ModelOutput:
        """
        Apply the encoder to the input.

        :param piece_ids:
            Piece identifiers to apply the encoder to.

            *Shape:* ``(batch_size, seq_len)``
        :param attention_mask:
            Attention mask. Sequence elements for which the
            corresponding mask element is set to ``False`` are ignored
            during attention calculation.
        :param positions:
            Input positions. Positions are used to look up position embeddings.
            Normally, these positions are calculated automatically. But if the
            positions deviate for some reason, they can be provided through this argument.

            *Shape:* ``(batch_size, seq_len)``
        :param type_ids:
            Type identifiers to indicate the spans of different
            sequences in the input. Useful when performing tasks like
            sequence classification and question answering.

            *Shape:* ``(batch_size, seq_len)``
        :returns:
            Encoder output.
        """
        raise NotImplementedError



# Copyright 2020 The HuggingFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# These are functions from Huggingface transformers. They are needed
# to get full output compatibility.


# Source:
# https://github.com/huggingface/transformers/blob/239ace152bfa84188d78d3802d2e69b64752be7e/src/transformers/tokenization_utils_base.py#L3578
def clean_up_decoded_string_like_hf(out_string: str) -> str:
    """
    Clean up a list of simple English tokenization artifacts like spaces before punctuations and abbreviated forms.

    :param out_string:
        The text to clean up.
    :returns:
        The cleaned-up string.
    """
    out_string = (
        out_string.replace(" .", ".")
        .replace(" ?", "?")
        .replace(" !", "!")
        .replace(" ,", ",")
        .replace(" ' ", "'")
        .replace(" n't", "n't")
        .replace(" 'm", "'m")
        .replace(" 's", "'s")
        .replace(" 've", "'ve")
        .replace(" 're", "'re")
    )
    return out_string


def tokenize_chinese_chars_bert(text: str) -> str:
    """Adds whitespace around any CJK character."""
    output = []
    for char in text:
        cp = ord(char)
        if _is_chinese_char_bert(cp):
            output.append(" ")
            output.append(char)
            output.append(" ")
        else:
            output.append(char)
    return "".join(output)


def _is_chinese_char_bert(cp: int) -> bool:
    """Checks whether CP is the codepoint of a CJK character."""
    # This defines a "chinese character" as anything in the CJK Unicode block:
    #   https://en.wikipedia.org/wiki/CJK_Unified_Ideographs_(Unicode_block)
    #
    # Note that the CJK Unicode block is NOT all Japanese and Korean characters,
    # despite its name. The modern Korean Hangul alphabet is a different block,
    # as is Japanese Hiragana and Katakana. Those alphabets are used to write
    # space-separated words, so they are not treated specially and handled
    # like the all of the other languages.
    if (
        (cp >= 0x4E00 and cp <= 0x9FFF)
        or (cp >= 0x3400 and cp <= 0x4DBF)  #
        or (cp >= 0x20000 and cp <= 0x2A6DF)  #
        or (cp >= 0x2A700 and cp <= 0x2B73F)  #
        or (cp >= 0x2B740 and cp <= 0x2B81F)  #
        or (cp >= 0x2B820 and cp <= 0x2CEAF)  #
        or (cp >= 0xF900 and cp <= 0xFAFF)
        or (cp >= 0x2F800 and cp <= 0x2FA1F)  #
    ):  #
        return True

    return False



from typing import Any, Dict, Mapping, Optional, Set, Tuple, Type, TypeVar

import torch
from torch import Tensor
from torch.nn import Linear

from ...quantization.quantizable import Quantizable
from ..hf_hub import FromHF
from ..hf_hub.conversion import state_dict_from_hf, state_dict_to_hf
from ..transformer import TransformerCausalLM
from ._hf import CAUSAL_LM_HF_PARAM_KEY_TRANSFORMS, _config_from_hf, _config_to_hf
from .config import FalconConfig
from .decoder import FalconDecoder

# Only provided as typing.Self in Python 3.11+.
Self = TypeVar("Self", bound="FalconCausalLM")


class FalconCausalLM(
    TransformerCausalLM[FalconConfig], FromHF[FalconConfig], Quantizable
):
    """
    Falcon (`Penedo et al., 2019`_) causal language model.

    .. _Penedo et al., 2019: https://arxiv.org/abs/2306.01116
    """

    def __init__(
        self, config: FalconConfig, *, device: Optional[torch.device] = None
    ) -> None:
        """
        Construct a Falcon causal LM.

        :param config:
            Causal LM configuration.
        :param device:
            Device to which the module is to be moved.
        :returns:
            The causal LM.
        """
        super().__init__(config)

        self.decoder = FalconDecoder(config, device=device)
        self.output_embeddings = Linear(
            in_features=config.layer.feedforward.hidden_width,
            out_features=config.embedding.n_pieces,
            bias=False,
            device=device,
        )

    @classmethod
    def state_dict_from_hf(
        cls: Type[Self], params: Mapping[str, Tensor]
    ) -> Mapping[str, Tensor]:
        return state_dict_from_hf(params, CAUSAL_LM_HF_PARAM_KEY_TRANSFORMS)

    @classmethod
    def is_supported(cls: Type[Self], config: Dict[str, Any]) -> bool:
        return config.get("model_type") in ("falcon", "RefinedWeb", "RefinedWebModel")

    @classmethod
    def state_dict_to_hf(
        cls: Type[Self], params: Mapping[str, Tensor]
    ) -> Mapping[str, Tensor]:
        return state_dict_to_hf(params, CAUSAL_LM_HF_PARAM_KEY_TRANSFORMS)

    @classmethod
    def config_from_hf(cls, hf_config: Mapping[str, Any]) -> FalconConfig:
        return _config_from_hf(hf_config)

    @classmethod
    def config_to_hf(cls, curated_config: FalconConfig) -> Mapping[str, Any]:
        return _config_to_hf(cls, curated_config)

    @classmethod
    def from_hf_config(
        cls: Type[Self],
        *,
        hf_config: Any,
        device: Optional[torch.device] = None,
    ) -> Self:
        config = cls.config_from_hf(hf_config)
        return cls(config, device=device)

    @classmethod
    def modules_to_not_quantize(cls) -> Set[str]:
        # Ignore the output embedding matrix.
        return {"output_embeddings"}



from typing import Generic, Iterable, List

from ..models.output import CacheT
from ..tokenizers.chunks import InputChunks
from ..tokenizers.tokenizer import TokenizerBase
from .config import GeneratorConfig
from .generator import Generator


class StringGenerator(Generic[CacheT]):
    """
    Generator wrapper that takes textual input and outputs generated strings.
    It wraps a generator and uses a tokenizer to split the input into pieces
    and decode the output pieces.
    """

    inner: Generator[CacheT]
    tokenizer: TokenizerBase

    def __init__(self, tokenizer: TokenizerBase, generator: Generator[CacheT]) -> None:
        """
        Construct a string generator.

        :param tokenizer:
            Tokenizer for piece processing.
        :param generator:
            Generator to wrap.
        """
        self.inner = generator
        self.tokenizer = tokenizer

    def __call__(
        self, prompts: Iterable[InputChunks], config: GeneratorConfig
    ) -> List[str]:
        """
        Alias for :meth:`.generate`.
        """
        return self.generate(prompts, config=config)

    def generate(
        self, prompts: Iterable[InputChunks], config: GeneratorConfig
    ) -> List[str]:
        """
        Generate text using the given prompts.

        :param prompts:
            Prompts to generate from.
        :param config:
            Generator configuraton.
        :returns:
            Strings generated for the prompts.
        """

        device = next(self.inner.model.parameters()).device
        pieces = self.tokenizer(prompts)
        ids = pieces.padded_tensor(pad_left=True, device=device)
        attention_mask = pieces.attention_mask(pad_left=True, device=device)

        piece_ids: List[List[int]] = [[] for _ in range(ids.size(0))]
        for seq_ids, outputs in self.inner(
            ids=ids, attention_mask=attention_mask, config=config
        ):
            for seq_id, seq_piece_ids in zip(seq_ids.tolist(), outputs.tolist()):
                piece_ids[seq_id].extend(seq_piece_ids)

        return self.tokenizer.decode(piece_ids)
